from __future__ import annotations
import math
from functools import lru_cache

import lazy_loader as lazy # type: ignore

from .pixelscale import PixelScale

# Pyproj is relatively slow to import, which was adding reasonably to
# the import time for yirgacheffe, so make it lazy
pyproj = lazy.load('pyproj')

# As per https://xkcd.com/2170/, we need to stop caring about floating point
# accuracy at some point as it becomes problematic.
# The value here is 1 meter, given that geo data that we've been working with
# is accurate to 100 meter, but if you need to worry about the biodiversity
# of a virus in a petri dish this assumption may not work for you.
#
# TODO: all these values are very much picked due to the global projects Yirgacheffe was
# originally developed as part of. Given we now see satellite image data down to tens of
# centimetres being procured, these values are questionable, and should probably be a
# fraction of the actual pixel scale.
MINIMAL_DISTANCE_OF_INTEREST_IN_M = 1.0
DISTANCE_PER_DEGREE_AT_EQUATOR = 40075017 / 360
MINIMAL_DEGREE_OF_INTEREST = MINIMAL_DISTANCE_OF_INTEREST_IN_M / DISTANCE_PER_DEGREE_AT_EQUATOR

# It turns out that calls to CRS.to_epsg are quite slow - not a problem as a one
# of call, but in our unit tests we make and destroy thousands of rasters, each of
# which has to do a to_epsg call to get the projection for saving. We saw one test
# that makes 1600 rasters take 72 seconds without this fix, and 2 seconds with it.
@lru_cache(maxsize=128)
def _get_projection_string(provided_name: str) -> str:
    crs = pyproj.CRS.from_string(provided_name)
    epsg = crs.to_epsg()
    if epsg is not None:
        return f"EPSG:{epsg}"
    return crs.to_wkt()

class MapProjection:
    """Records the map projection and the size of the pixels in a layer.

    Note: It is very common to find that round errors creep into pixel scale values in
    GeoTIFFs from different sources, and so MapProjection tolerates small amounts of difference between
    pixel scales that are below a single metre in resolution.

    This superceeeds the old PixelScale class, which will be removed in version 2.0.

    Args:
        name: The map projection used in WKT format, or as "epsg:xxxx" or "esri:xxxx".
        xstep: The number of units horizontal distance a step of one pixel makes in the map projection.
        ystep: The number of units vertical distance a step of one pixel makes in the map projection.

    Attributes:
        name: The map projection used in WKT format.
        xstep: The number of units horizontal distance a step of one pixel makes in the map projection.
        ystep: The number of units vertical distance a step of one pixel makes in the map projection.

    Examples:
        Create a map projection using an EPSG code:

        >>> proj_wgs84 = MapProjection("epsg:4326", 0.001, -0.001)

        Create a projection using an ESRI code:

        >>> proj_esri = MapProjection("esri:54030", 1000, -1000)
    """

    def __init__(self, projection_string: str, xstep: float, ystep: float) -> None:
        try:
            self.crs = pyproj.CRS.from_string(projection_string)
        except pyproj.exceptions.CRSError as exc:
            raise ValueError(f"Invalid projection: {projection_string}") from exc
        self.xstep = xstep
        self.ystep = ystep
        self._gdal_projection = _get_projection_string(projection_string)

    @property
    def _min_step(self) -> float:
        unit_name = self.crs.axis_info[0].unit_name
        if unit_name in ('metre', 'meter', 'm'):
            return MINIMAL_DISTANCE_OF_INTEREST_IN_M
        elif unit_name in ('degree', 'degrees'):
            return MINIMAL_DEGREE_OF_INTEREST
        else:
            raise NotImplementedError(f"Unsupported unit: {unit_name}")

    def __repr__(self) -> str:
        return f"MapProjection({self.crs.to_string()!r}, {self.xstep}, {self.ystep})"

    def __hash__(self):
        return hash((self.name, self.xstep, self.ystep))

    def __eq__(self, other) -> bool:
        if not isinstance(other, MapProjection):
            return False
        if self.crs != other.crs:
            return False
        return (abs(self.xstep - other.xstep) < self._min_step) and \
            (abs(self.ystep - other.ystep) < self._min_step)

    @property
    def name(self) -> str:
        return self.crs.to_wkt()

    @property
    def epsg(self) -> int | None:
        return self.crs.to_epsg()

    @property
    def scale(self) -> PixelScale:
        return PixelScale(self.xstep, self.ystep)

    def round_up_pixels(self, x: float, y: float) -> tuple[int, int]:
        """In general we round up pixels, as we don't want to lose range data,
        but floating point math means we will get errors where the value of pixel
        scale value rounds us to a tiny faction of a pixel up, and so math.ceil
        would round us up for microns worth of distance. """
        floored_x = math.floor(x)
        floored_y = math.floor(y)
        diff_x = x - floored_x
        diff_y = y - floored_y
        spatial_diff_x = diff_x * abs(self.xstep)
        spatial_diff_y = diff_y * abs(self.ystep)

        final_x = floored_x if (spatial_diff_x < self._min_step) else math.ceil(x)
        final_y = floored_y if (spatial_diff_y < self._min_step) else math.ceil(y)

        return (final_x, final_y)

    def round_down_pixels(self, x: float, y: float) -> tuple[int, int]:
        ceiled_x = math.ceil(x)
        ceiled_y = math.ceil(y)
        diff_x = ceiled_x - x
        diff_y = ceiled_y - y
        spatial_diff_x = diff_x * abs(self.xstep)
        spatial_diff_y = diff_y * abs(self.ystep)

        final_x = ceiled_x if (spatial_diff_x < self._min_step) else math.floor(x)
        final_y = ceiled_y if (spatial_diff_y < self._min_step) else math.floor(y)

        return (final_x, final_y)



from __future__ import annotations
import math
from typing import Any

import numpy as np
from pyproj import Transformer

from .._datatypes import Area, MapProjection, Window
from .base import YirgacheffeLayer
from .._backends import backend
from .._backends.enumeration import dtype as DataType

# Based on https://gis.stackexchange.com/questions/127165/more-accurate-way-to-calculate-area-of-rasters
def area_of_pixel(a: float, b: float, pixel_scale: tuple[float,float], center_lat: float) -> float:
    x_scale, y_scale = pixel_scale
    e = math.sqrt(1 - ((b / a) ** 2))
    area_list = []
    for f in [center_lat + (y_scale / 2), center_lat - (y_scale / 2)]:
        sin_of_f = math.sin(math.radians(f))
        zm = 1 - (e * sin_of_f)
        zp = 1 + (e * sin_of_f)
        area_list.append(
            math.pi * (b ** 2) * (
                (math.log(zp / zm) / (2 * e)) +
                (sin_of_f / (zp * zm))
            )
        )
    return abs((x_scale / 360.) * (area_list[0] - area_list[1]))

class AreaPerPixelLayer(YirgacheffeLayer):
    """This layer will dynamically generate pixels where the value is the area in square meters of that pixel in
    the given map projection and pixel scale.
    """

    def __init__(self, projection: MapProjection):
        if projection is None:
            raise ValueError("projection should not be None")
        if not isinstance(projection, MapProjection):
            raise TypeError("projection should be of type MapProjection")

        if projection.crs.area_of_use is not None:
            west, south, east, north = projection.crs.area_of_use.bounds
            name = projection.crs.area_of_use.name
        else:
            # assume a global bounds
            west, south, east, north = -180.0, -90.0, 180.0, 90.0
            name = projection.crs.name

        x_scale = abs(projection.xstep)
        y_scale = abs(projection.ystep)

        if projection.crs.is_geographic:
            area = Area(
                left=math.floor(west / x_scale) * x_scale,
                top=math.ceil(north / y_scale) * y_scale,
                right=math.ceil(east / x_scale) * x_scale,
                bottom=math.floor(south / y_scale) * y_scale,
                projection=projection,
            )
        else:
            # For projections like Mollweide, the corners aren't valid extremes, so we test
            # a mix of corners and mid-points and take the extremes on each axis
            horizontal = [west, ((west + east) / 2), east]
            vertical = [north, ((south + north) / 2), south]

            points = []
            for x in horizontal:
                points.extend([(x, i) for i in vertical])

            transformer = Transformer.from_crs("epsg:4326", projection.crs, always_xy=True)
            xs, ys = transformer.transform(*zip(*points)) # type: ignore

            left = math.floor(min(xs) / x_scale) * x_scale
            right = math.ceil(max(xs) / x_scale) * x_scale
            bottom = math.floor(min(ys) / y_scale) * y_scale
            top = math.floor(max(ys) / y_scale) * y_scale
            area = Area(left=left, top=top, right=right, bottom=bottom, projection=projection)

        super().__init__(area, name)

        self._hash = hash(projection._gdal_projection)

    @property
    def _cse_hash(self) -> int | None:
        return self._hash

    @property
    def datatype(self):
        return DataType.Float32

    def _read_array_with_window(
        self,
        _xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:
        projection = self.map_projection
        if projection is None:
            raise RuntimeError("Area Per Pixel layer should always be projected")

        if projection.crs.is_geographic:
            ellipsoid = projection.crs.ellipsoid
            if ellipsoid is None:
                raise RuntimeError("Exected geographic CRS to have ellipsoid details")
            a = ellipsoid.semi_major_metre
            b = ellipsoid.semi_minor_metre

            offset = window.yoff + yoffset
            x_scale = projection.xstep
            y_scale = projection.ystep

            area_values = np.array([
                area_of_pixel(a, b, (x_scale, y_scale), self._underlying_area.top + ((offset + i + 0.5) * y_scale))
                for i in range(ysize)
            ])
            area_array = np.broadcast_to(area_values[:, np.newaxis], (ysize, xsize))
            return backend.promote(area_array)

        else:
            pixel_area = abs(projection.xstep * projection.ystep)
            return backend.full((ysize, xsize), pixel_area)



from __future__ import annotations
import math
from dataclasses import dataclass

from osgeo import gdal

from .mapprojection import MapProjection

@dataclass(frozen=True)
class Area:
    """Class to hold a geospatial area. Can optionally have a projection associated.

    Ideally areas should always have a projection associated with them, however some data sources,
    notably polygon datasets like GeoJSON, do not store this, so we have to allow for projectionless
    areas.

    You can use set operators | (union) and & (intersection) on Areas.

    If two areas are intersected or unioned and they have the same map projection and geospatial
    pixel size, but they do not perfectly align on the same pixel grid, then the operation will be
    performed based on nearest neighbour alignment of the pixel grids, such that the resulting area
    is still pixel aligned.

    Args:
        left: Left most point in the projection space.
        top: Top most point in the projection space.
        right: Right most point in the projection space.
        bottom: Bottom most point in the projection space.
        projection: An optional map projection.

    Attributes:
        left: Left most point in the projection space.
        top: Top most point in the projection space.
        right: Right most point in the projection space.
        bottom: Bottom most point in the projection space.
        projection: An optional map projection.
    """
    left: float
    top: float
    right: float
    bottom: float
    projection: MapProjection | None = None

    def __post_init__(self) -> None:
        if self.projection is None:
            return

        # If we have a map projection then validate the dimensions are based on the pixel scale
        # Note we have to not use the obvious approach of using mod here, as floating point mod
        # is a bit iffy with values between 1 and 0
        # >>> 20 % 2
        # 0
        # >>> 20 % 1
        # 0
        # >>> 20 % 0.1
        # 0.0999999999999989
        # Note this is an FP issue not a Python issue AFAICT, I get the same behaviour in OCaml
        # for instance.
        width = self.right - self.left
        height = self.top - self.bottom
        x_pixels = abs(width / self.projection.xstep)
        y_pixels = abs(height / self.projection.ystep)

        if not math.isclose(x_pixels - round(x_pixels), 0.0, abs_tol=1e-09) or \
                not math.isclose(y_pixels - round(y_pixels), 0.0, abs_tol=1e-09):
            raise ValueError("Area expected to be an integer multiple of projection units")

    @staticmethod
    def world() -> Area:
        """Creates an area that covers the entire planet.

        Returns:
            An area where the extents are nan, but is_world returns true.
        """
        return Area(float("nan"), float("nan"), float("nan"), float("nan"))

    def __hash__(self):
        return (self.left, self.top, self.right, self.bottom).__hash__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Area):
            return False
        if self.is_world and other.is_world:
            return True
        if self.projection != other.projection:
            return False

        self_offset = self._grid_offset
        other_offset = other._grid_offset

        if self_offset and other_offset:
            x_diff = self_offset[0] - other_offset[0]
            y_diff = self_offset[1] - other_offset[1]
        else:
            x_diff = 0.0
            y_diff = 0.0

        return \
            math.isclose(self.left, other.left + x_diff, abs_tol=1e-09) and \
            math.isclose(self.right, other.right + x_diff, abs_tol=1e-09) and \
            math.isclose(self.top, other.top + y_diff, abs_tol=1e-09) and \
            math.isclose(self.bottom, other.bottom + y_diff, abs_tol=1e-09)

    def __and__(self, other: object) -> Area:
        # Set intersection
        if not isinstance(other, Area):
            raise  ValueError("Can only intersect two areas")
        if self.is_world:
            return other
        if other.is_world:
            return self

        lhs = self
        rhs = other

        if lhs.projection is None and rhs.projection is not None:
            lhs = lhs.project_like(rhs)
        elif rhs.projection is None and lhs.projection is not None:
            rhs = rhs.project_like(lhs)

        if lhs.projection != rhs.projection:
            raise ValueError("Cannot intersect areas with different projections")

        # If we intersect two layers with different grid wobbles, then generate
        # a result that is aligned with the midpoint between them.
        lhs_offset = lhs._grid_offset
        rhs_offset = rhs._grid_offset

        if lhs_offset and rhs_offset:
            x_offset = (lhs_offset[0] + rhs_offset[0]) / 2
            y_offset = (lhs_offset[1] + rhs_offset[1]) / 2
        else:
            lhs_offset = (0.0, 0.0)
            rhs_offset = (0.0, 0.0)
            x_offset = 0.0
            y_offset = 0.0

        # We do this on the grid aligned space, as it's easier to spot bugs in the
        # numbers that way (usually)
        intersection = Area(
            left=max(lhs.left - lhs_offset[0], rhs.left - rhs_offset[0]),
            top=min(lhs.top - lhs_offset[1], rhs.top - rhs_offset[1]),
            right=min(lhs.right - lhs_offset[0], rhs.right - rhs_offset[0]),
            bottom=max(lhs.bottom - lhs_offset[1], rhs.bottom - rhs_offset[1]),
            projection=lhs.projection,
        )

        if (intersection.left >= intersection.right) or (intersection.bottom >= intersection.top) or \
                 math.isclose(intersection.left, intersection.right) or \
                 math.isclose(intersection.top, intersection.bottom):
            raise ValueError('No intersection possible')
        return Area(
            intersection.left + x_offset,
            intersection.top + y_offset,
            intersection.right + x_offset,
            intersection.bottom + y_offset,
            intersection.projection,
        )

    def __or__(self, other: object) -> Area:
        # Set union
        if not isinstance(other, Area):
            raise  ValueError("Can only union two areas")
        if self.is_world:
            return self
        if other.is_world:
            return other

        lhs = self
        rhs = other

        if lhs.projection is None and rhs.projection is not None:
            lhs = lhs.project_like(rhs)
        elif rhs.projection is None and lhs.projection is not None:
            rhs = rhs.project_like(lhs)

        if lhs.projection != rhs.projection:
            raise ValueError("Cannot union areas with different projections")

        # If we union two layers with different grid wobbles, then generate
        # a result that is aligned with the midpoint between them.
        lhs_offset = lhs._grid_offset
        rhs_offset = rhs._grid_offset
        if lhs_offset and rhs_offset:
            x_offset = (lhs_offset[0] + rhs_offset[0]) / 2
            y_offset = (lhs_offset[1] + rhs_offset[1]) / 2
        else:
            lhs_offset = (0.0, 0.0)
            rhs_offset = (0.0, 0.0)
            x_offset = 0.0
            y_offset = 0.0

        union = Area(
            left=min(lhs.left - lhs_offset[0], rhs.left - rhs_offset[0]) + x_offset,
            top=max(lhs.top - lhs_offset[1], rhs.top - rhs_offset[1]) + y_offset,
            right=max(lhs.right - lhs_offset[0], rhs.right - rhs_offset[0]) + x_offset,
            bottom=min(lhs.bottom - lhs_offset[1], rhs.bottom - rhs_offset[1]) + y_offset,
            projection=lhs.projection,
        )
        return union

    @property
    def _grid_offset(self) -> tuple[float, float] | None:
        if self.projection is None:
            return None

        abs_xstep = abs(self.projection.xstep)
        abs_ystep = abs(self.projection.ystep)

        xoff = self.left % abs_xstep
        yoff = self.top % abs_ystep

        # We need to be consistent about what happens when
        # layers align by half a cell, and so we always nudge
        # things by a fractional pixel area to ensure this.
        # Otherwise the usual >= 0.5 wobbles due to rounding
        # errors (ask me how I know...).
        epsilon_x = abs_xstep * 1e-6
        epsilon_y = abs_ystep * 1e-6

        threshold_x = abs_xstep / 2
        threshold_y = abs_ystep / 2

        if xoff > threshold_x + epsilon_x:
            xoff -= abs_xstep
        elif xoff > threshold_x - epsilon_x:
            xoff = threshold_x

        if yoff > threshold_y + epsilon_y:
            yoff -= abs_ystep
        elif yoff > threshold_y - epsilon_y:
            yoff = threshold_y

        return xoff, yoff

    @property
    def _grid_aligned(self) -> Area:
        offset = self._grid_offset
        if offset is None:
            return self
        return Area(
            self.left - offset[0],
            self.top - offset[1],
            self.right - offset[0],
            self.bottom - offset[1],
            self.projection,
        )

    def grow(self, offset: float) -> Area:
        """Expand the area in all directions by the given amount.

        Generates a new area that is an expanded version of the current area.

        Args:
            offset: The amount by which to grow the area.

        Returns:
            A new area of the expanded size.
        """
        return Area(
            left=self.left - offset,
            top=self.top + offset,
            right=self.right + offset,
            bottom=self.bottom - offset,
            projection=self.projection,
        )

    @property
    def is_world(self) -> bool:
        """Returns true if this is a global area, independent of projection.

        Returns:
            True if the Area was created with `world` otherwise False.
        """
        return math.isnan(self.left)

    def overlaps(self, other: Area) -> bool:
        """Check if this area overlaps with another area.

        Args:
            other: The other area to compare this area with.

        Returns:
            True if the two areas intersect, otherwise false.
        """

        if self.is_world or other.is_world:
            return True

        lhs = self
        rhs = other

        if lhs.projection is None and rhs.projection is not None:
            lhs = lhs.project_like(rhs)
        elif rhs.projection is None and lhs.projection is not None:
            rhs = rhs.project_like(lhs)

        if lhs.projection != rhs.projection:
            raise ValueError("Cannot compare areas with different projections")

        return (
            (lhs.left <= rhs.left <= lhs.right) or
            (lhs.left <= rhs.right <= lhs.right) or
            (rhs.left <= lhs.left <= rhs.right) or
            (rhs.left <= lhs.right <= rhs.right)
        ) and (
            (lhs.top >= rhs.top >= lhs.bottom) or
            (lhs.top >= rhs.bottom >= lhs.bottom) or
            (rhs.top >= lhs.top >= rhs.bottom) or
            (rhs.top >= lhs.bottom >= rhs.bottom)
        )

    def reproject(self, target_projection: MapProjection) -> Area:
        """Takes an area and projects it."""
        if target_projection is None:
            raise ValueError("Target projection can not be None")
        if self.projection is None:
            # This is because if you're going from a unprojected area, which will be a VectorLayer at the
            # time of writing this comment, the target needs to be a raster with a grid offset to align to
            # so needs an Area and not a MapProjection input.
            raise ValueError("Can only reproject already projected areas. Use `project_like` instead.")

        # Converting between projections is tricky, as some projections either skew or warp as translated
        # and so to avoid implementing a lot of complex math here, I'll just ask GDAL what it would do
        # if warping from our projection to the other.
        width, height = self.pixel_dimensions

        # Create a VRT (just XML, no data allocation)
        vrt_xml = f'''<VRTDataset rasterXSize="{width}" rasterYSize="{height}">
          <SRS>{self.projection.crs.to_wkt()}</SRS>
          <GeoTransform>{self.left}, {self.projection.xstep}, 0,
          {self.top}, 0, {self.projection.ystep}</GeoTransform>
          <VRTRasterBand dataType="Byte" band="1"/>
        </VRTDataset>'''

        src_ds = gdal.Open(vrt_xml)
        vrt_ds = gdal.AutoCreateWarpedVRT(src_ds, None,
                                          target_projection._gdal_projection,
                                          gdal.GRA_NearestNeighbour)
        gt = vrt_ds.GetGeoTransform()
        min_x = gt[0]
        max_y = gt[3]
        max_x = min_x + gt[1] * vrt_ds.RasterXSize
        min_y = max_y + gt[5] * vrt_ds.RasterYSize

        return Area(
            math.floor(min_x / target_projection.xstep) * target_projection.xstep,
            math.ceil(max_y / target_projection.ystep) * target_projection.ystep,
            math.ceil(max_x / target_projection.xstep) * target_projection.xstep,
            math.floor(min_y / target_projection.ystep) * target_projection.ystep,
            target_projection,
        )

    def project_like(self, other: Area) -> Area:
        """Takes a projectionless area and maps it onto a map projection based on an existing area.

        Because map projections have pixel scales associated with them, the area may be expanded
        to ensure that the original area is within the bounds when mapped to the pixel space of the other area.

        Will raise an exception if this area already has a map projection set, or if the other area does not.

        Args:
            other: The other area to take the map projection from.

        Returns:
            A new area with the projection map.
        """
        if other.projection is None:
            raise ValueError("Like area must be have map projection")
        if self.projection is None:
            offset = other._grid_offset
            assert offset # We know this should be true due to the above guard, but pylint does not.
            x_off, y_off = offset
            abs_xstep = abs(other.projection.xstep)
            abs_ystep = abs(other.projection.ystep)

            return Area(
                left=(math.floor(self.left / abs_xstep) * abs_xstep) + x_off,
                top=(math.ceil(self.top / abs_ystep) * abs_ystep) + y_off,
                right=(math.ceil(self.right / abs_xstep) * abs_xstep) + x_off,
                bottom=(math.floor(self.bottom / abs_ystep) * abs_ystep) + y_off,
                projection=other.projection
            )
        else:
            # This is because I should have always had two entry points, but briefly
            # in 1.11.3 I didn't foresee that. This should be tidied up in the 2.0 API
            # clean up.
            return self.reproject(other.projection)

    @property
    def pixel_dimensions(self) -> tuple[int,int]:
        """Returns the size in pixels for this area in its given projection.

        Attempts to call this on an area with no projection will raise a ValueError.

        Returns:
            A tuple of the width and height.
        """
        if self.projection is None:
            raise ValueError("No dimensions for unprojected area")
        abs_xstep, abs_ystep = abs(self.projection.xstep), abs(self.projection.ystep)
        return self.projection.round_up_pixels(
            (self.right - self.left) / abs_xstep,
            (self.top - self.bottom) / abs_ystep,
        )

    @property
    def geo_transform(self) -> tuple[float,float,float,float,float,float]:
        """Returns the GDAL geo transform for the area.__and__()

        Attempts to call this on an area with no projection will raise a ValueError.

        Returns:
            A tuple of floats for the GDAL geo transform record.
        """
        if self.projection is None:
            raise ValueError("No geo transform for unprojected area")
        return (
            self.left,
            self.projection.xstep,
            0.0,
            self.top,
            0.0,
            self.projection.ystep
        )



# this code has moved, and in 2.0 this file should be deleted and the items accessed via the top level import
# pylint: disable=W0611
from ._datatypes.area import Area
from ._datatypes.mapprojection import MapProjection, PixelScale
from ._datatypes.window import Window



import operator as pyoperator

from . import LayerOperation, WindowOperation
from .._backends.enumeration import operators as op

def where(cond, a, b):
    """Return elements chosen from `a` or `b` depending on `cond`.

    Behaves like numpy.where(condition, x, y), returning a layer operation
    where elements from `a` are selected where `cond` is True, and elements
    from `b` are selected where `cond` is False.

    Args:
        cond: Layer or constant used as condition. Where True, yield `a`, otherwise yield `b`.
        a: Layer or constant with values from which to choose where `cond` is True.
        b: Layer or constant with values from which to choose where `cond` is False.

    Returns:
        New layer representing the conditional selection.
    """
    return LayerOperation(
        cond,
        op.WHERE,
        rhs=a,
        other=b
    )

def maximum(a, b):
    """Element-wise maximum of layer elements.

    Behaves like numpy.maximum(x1, x2), comparing two layers element-by-element
    and returning a new layer with the maximum values.

    Args:
        a: First layer or constant to compare.
        b: Second layer or constant to compare.

    Returns:
        New layer representing the element-wise maximum of the inputs.
    """
    return LayerOperation(
        a,
        op.MAXIMUM,
        b,
        window_op=WindowOperation.UNION,
    )

def minimum(a, b):
    """Element-wise minimum of layer elements.

    Behaves like numpy.minimum(x1, x2), comparing two layers element-by-element
    and returning a new layer with the minimum values.

    Args:
        a: First layer or constant to compare.
        b: Second layer or constant to compare.

    Returns:
        New layer representing the element-wise minimum of the inputs.
    """
    return LayerOperation(
        a,
        op.MINIMUM,
        rhs=b,
        window_op=WindowOperation.UNION,
    )

def _balanced_reduce(layers, operator):
    if len(layers) == 0:
        raise RuntimeError("Internal precondition violation")
    if len(layers) == 1:
        return layers[0]
    mid = len(layers) // 2
    left = _balanced_reduce(layers[:mid], operator)
    right = _balanced_reduce(layers[mid:], operator)
    return operator(left, right)

def sum(layers: list): # pylint: disable=W0622
    """Combine multiple layers by summing spatially corresponding pixels.

    Creates a new raster where each pixel is the sum of that pixel's
    values across all input rasters for the same location.

    Args:
        layers: List/sequence of layers to sum

    Returns:
        A new raster layer with pixel-wise sums

    Examples:
        # Combine 100 species habitat rasters into richness map
        richness = yg.sum(habitat_layers)

    Note:
        To sum all pixels within a single raster to get a scalar,
        use the .sum() method instead: `layer.sum()`
    """
    if len(layers) == 0:
        raise ValueError("List of layers is empty")
    return _balanced_reduce(layers, pyoperator.__add__)

def any(layers: list): # pylint: disable=W0622
    """Combine layers with OR operation.

    Returns a raster where each pixel is 1 if ANY input layers
    has a non-zero value at that geospatial location.

    Similar to Python's built-in any() but for Yirgacheffe layers.

    Args:
        layers: List/sequence of layers to OR

    Returns:
        A new raster layer with pixel-wise sums
    """
    if len(layers) == 0:
        raise ValueError("List of layers is empty")
    return _balanced_reduce([layer != 0 for layer in layers], pyoperator.or_)

def all(layers: list): # pylint: disable=W0622
    """Combine layers with AND operation.

    Returns a raster where each pixel is 1 only if ALL input layers
    have non-zero values at that geospatial location.

    Similar to Python's built-in all() but for Yirgacheffe layers.

    Args:
        layers: List/sequence of layers to AND

    Returns:
        A new raster layer with pixel-wise sums
    """
    if len(layers) == 0:
        raise ValueError("List of layers is empty")
    return _balanced_reduce([layer != 0 for layer in layers], pyoperator.and_)

# We provide these module level accessors as it's often nicer to write `log(x/y)` rather than `(x/y).log()`
clip = LayerOperation.clip
log = LayerOperation.log
log2 = LayerOperation.log2
log10 = LayerOperation.log10
exp = LayerOperation.exp
exp2 = LayerOperation.exp2
nan_to_num = LayerOperation.nan_to_num
isin = LayerOperation.isin
abs = LayerOperation.abs # pylint: disable=W0622
floor = LayerOperation.floor
round = LayerOperation.round # pylint: disable=W0622
ceil = LayerOperation.ceil



from pathlib import Path

from osgeo import gdal
import tomli as tomllib

try:
    from importlib import metadata
    __version__: str = metadata.version(__name__)
except ModuleNotFoundError:
    pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
    with open(pyproject_path, "rb") as f:
        pyproject_data = tomllib.load(f)
    __version__ = pyproject_data["project"]["version"]

from .layers import YirgacheffeLayer, ResamplingMethod
from ._core import read_raster, read_rasters, read_shape, read_raster_like, read_shape_like, constant, \
    read_narrow_raster, from_array, area_raster, to_geotiff
from .constants import WGS_84_PROJECTION
from ._datatypes.area import Area
from ._datatypes.mapprojection import MapProjection
from ._datatypes.pixelscale import PixelScale
from ._datatypes.window import Window
from ._backends.enumeration import dtype as DataType

from ._operators.functions import where, minimum, maximum, clip, log, log2, log10, exp, exp2, nan_to_num, isin, \
    floor, ceil # pylint: disable=W0611
from ._operators.functions import abs, all, any, round, sum # pylint: disable=W0611,W0622

gdal.UseExceptions()



from __future__ import annotations
import operator
import uuid
from collections.abc import Callable
from functools import reduce
from typing import Any, Sequence

import deprecation

from .. import __version__
from .._operators import LayerMathMixin
from .._datatypes import Area, MapProjection, PixelScale, Window
from .._backends import backend
from .._backends.enumeration import dtype as DataType

class YirgacheffeLayer(LayerMathMixin):
    """The common base class for the different layer types. Most still inherit from RasterLayer as deep down
    they end up as pixels, but this is a start to make other layers that don't need to rasterize not have
    to carry all that baggage."""

    def __init__(self,
        area: Area,
        name: str | None = None
    ):
        self._underlying_area = area
        self._active_area: Area | None = None
        self._window: Window | None = None
        self.name = name if name is not None else str(uuid.uuid4())

        self.reset_window()

    def close(self) -> None:
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def _cse_hash(self) -> int | None:
        raise NotImplementedError("Must be overridden by subclass")

    def _park(self) -> None:
        pass

    def _unpark(self) -> None:
        pass

    @property
    def _raster_dimensions(self) -> tuple[int, int]:
        raise AttributeError("Does not have raster")

    @property
    def datatype(self) -> DataType:
        """Returns the [`DataType`][yirgacheffe.DataType] of the pixels within a layer."""
        raise NotImplementedError("Must be overridden by subclass")

    @property
    @deprecation.deprecated(
        deprecated_in="1.7",
        removed_in="2.0",
        current_version=__version__,
        details="Use `map_projection` instead."
    )
    def projection(self) -> str | None:
        if self.map_projection:
            return self.map_projection.name
        else:
            return None

    @property
    @deprecation.deprecated(
        deprecated_in="1.7",
        removed_in="2.0",
        current_version=__version__,
        details="Use `map_projection` instead."
    )
    def pixel_scale(self) -> PixelScale | None:
        if self.map_projection:
            return self.map_projection.scale
        else:
            return None

    @property
    def map_projection(self) -> MapProjection | None:
        """Returns the [`MapProjection`][yirgacheffe.MapProjection] (projection name and pixel size) of the layer."""
        return self._underlying_area.projection

    @property
    def area(self) -> Area:
        """Returns the geospatial [`Area`][yirgacheffe.Area] of the layer."""
        if self._active_area is not None:
            return self._active_area
        else:
            return self._underlying_area

    def _get_operation_area(
        self,
        projection: MapProjection | None = None,
        _force_union: bool = False,
        top_level: bool = False, # pylint: disable=W0613
    ) -> Area:
        self_projection = self.map_projection
        if self_projection is not None and projection is not None and self_projection != projection:
            raise ValueError("Calculation projection does not match layer projection")
        return self.area

    @property
    def window(self) -> Window:
        if self._window is None:
            raise AttributeError("Layer has no window")
        return self._window

    @property
    def nodata(self) -> None:
        return None

    @staticmethod
    def find_intersection(layers: Sequence[YirgacheffeLayer]) -> Area:
        if not layers:
            raise ValueError("Expected list of layers")

        # This only makes sense (currently) if all layers
        # have the same pixel pitch (modulo desired accuracy)
        projections = [x.map_projection for x in layers if x.map_projection is not None]
        if not projections:
            raise ValueError("No layers have a projection")
        if not all(projections[0] == x for x in projections[1:]):
            raise ValueError("Not all layers are at the same projection or pixel scale")

        layer_areas = [x._get_operation_area(projections[0]) for x in layers]
        return reduce(operator.and_, layer_areas)

    @staticmethod
    def find_union(layers: Sequence[YirgacheffeLayer]) -> Area:
        if not layers:
            raise ValueError("Expected list of layers")

        # This only makes sense (currently) if all layers
        # have the same pixel pitch (modulo desired accuracy)
        projections = [x.map_projection for x in layers if x.map_projection is not None]
        if not projections:
            raise ValueError("No layers have a projection")
        if not all(projections[0] == x for x in projections[1:]):
            raise ValueError("Not all layers are at the same projectin or pixel scale")

        layer_areas = [x._get_operation_area(projections[0]) for x in layers]
        # This removal of global layers is to stop constant layers forcing everything to be global
        return reduce(operator.or_, [x for x in layer_areas if not x.is_world])

    @property
    def geo_transform(self) -> tuple[float, float, float, float, float, float]:
        if self.map_projection is None:
            raise AttributeError("No geo transform for layers without explicit pixel scale")
        return (
            self.area.left, self.map_projection.xstep, 0.0,
            self.area.top, 0.0, self.map_projection.ystep
        )

    def _set_window(self, new_area: Area, validation_callable: Callable[[Window],None] | None = None) -> None:
        if self.map_projection is None:
            raise ValueError("Can not set Window without explicit pixel scale")

        if new_area.projection is None:
            new_area = new_area.project_like(self._underlying_area)

        # We force everything onto a grid aligned basis for calculating the window to avoid rounding issues
        new_area = new_area._grid_aligned
        underlying_area = self._underlying_area._grid_aligned

        xoff, yoff = self.map_projection.round_down_pixels(
            (new_area.left - underlying_area.left) / abs(self.map_projection.xstep),
            (underlying_area.top - new_area.top) / abs(self.map_projection.ystep),
        )
        xsize, ysize = self.map_projection.round_up_pixels(
            (new_area.right - new_area.left) / abs(self.map_projection.xstep),
            (new_area.top - new_area.bottom) / abs(self.map_projection.ystep),
        )
        new_window = Window(xoff, yoff, xsize, ysize)

        if validation_callable is not None:
            # This should raise an exception if it fails
            validation_callable(new_window)

        self._window = new_window
        self._active_area = new_area

    def set_window_for_intersection(self, new_area: Area) -> None:
        def _intersection_validation(new_window: Window) -> None:
            if (new_window.xoff < 0) or (new_window.yoff < 0):
                raise ValueError('Window has negative offset')
            # If there is an underlying raster for this layer, do a sanity check
            try:
                raster_xsize, raster_ysize = self._raster_dimensions
                if ((new_window.xoff + new_window.xsize) > raster_xsize) or \
                    ((new_window.yoff + new_window.ysize) > raster_ysize):
                    raise ValueError(f'Window is bigger than dataset: raster is {raster_xsize}x{raster_ysize}'\
                        f', new window is {new_window.xsize - new_window.xoff}x{new_window.ysize - new_window.yoff}')
            except AttributeError:
                pass
        self._set_window(new_area, _intersection_validation)

    def set_window_for_union(self, new_area: Area) -> None:
        def _union_validation(new_window: Window) -> None:
            if (new_window.xoff > 0) or (new_window.yoff > 0):
                raise ValueError('Window has positive offset')
            # If there is an underlying raster for this layer, do a sanity check
            try:
                raster_xsize, raster_ysize = self._raster_dimensions
                if ((new_window.xsize - new_window.xoff) < raster_xsize) or \
                    ((new_window.ysize - new_window.yoff) <raster_ysize):
                    raise ValueError(f'Window is smaller than dataset: raster is {raster_xsize}x{raster_ysize}'\
                        f', new window is {new_window.xsize - new_window.xoff}x{new_window.ysize - new_window.yoff}')
            except AttributeError:
                pass
        self._set_window(new_area, _union_validation)

    def reset_window(self) -> None:
        self._active_area = None
        if self.map_projection:
            width, height = self.map_projection.round_up_pixels(
                (self.area.right - self.area.left) / self.map_projection.xstep,
                (self.area.bottom - self.area.top) / self.map_projection.ystep,
            )
            self._window = Window(0, 0, width, height)

    def offset_window_by_pixels(self, offset: int) -> None:
        """Grows (if pixels is positive) or shrinks (if pixels is negative) the window for the layer."""
        if offset == 0:
            return

        if offset < 0:
            if (offset * -2 >= self.window.xsize) or (offset * -2 >= self.window.ysize):
                raise ValueError(f"Can not shrink window by {offset}, would make size 0 or fewer pixels.")

        new_window = Window(
            xoff=self.window.xoff - offset,
            yoff=self.window.yoff - offset,
            xsize=self.window.xsize + (2 * offset),
            ysize=self.window.ysize + (2 * offset),
        )
        scale = self.pixel_scale
        if scale is None:
            raise ValueError("Can not offset Window without explicit pixel scale")

        # Note we can't assume that we weren't already on an intersection when making the offset!
        # But remember that window is always relative to underlying area, and new_window
        # here is based off the existing window
        new_left = self._underlying_area.left + (new_window.xoff * scale.xstep)
        new_top = self._underlying_area.top + (new_window.yoff * scale.ystep)
        new_area = Area(
            left=new_left,
            top=new_top,
            right=new_left + (new_window.xsize * scale.xstep),
            bottom=new_top + (new_window.ysize * scale.ystep),
            projection=self._underlying_area.projection,
        )
        self._window = new_window
        self._active_area = new_area

    def _read_array_with_window(
        self,
        _x: int,
        _y: int,
        _xsize: int,
        _ysize: int,
        _window: Window,
    ) -> Any:
        raise NotImplementedError("Must be overridden by subclass")

    def _read_array_for_area(
        self,
        target_area: Area,
        target_projection: MapProjection,
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> Any:
        assert self.map_projection is not None
        assert self.map_projection == target_projection

        # move the target area to align with our grid offset
        target_offset = target_area._grid_offset
        if target_offset is not None:
            self_offset = self._underlying_area._grid_offset
            assert self_offset is not None
            target_offset_x, target_offset_y = target_offset
            self_offset_x, self_offset_y = self_offset
            diff_x = self_offset_x - target_offset_x
            diff_y = self_offset_y - target_offset_y
            target_area = Area(
                target_area.left + diff_x,
                target_area.top + diff_y,
                target_area.right + diff_x,
                target_area.bottom + diff_y,
                target_area.projection,
            )

        xoff, yoff = self.map_projection.round_down_pixels(
            (target_area.left - self._underlying_area.left) / self.map_projection.xstep,
            (self._underlying_area.top - target_area.top) / (self.map_projection.ystep * -1.0),
        )
        xsize, ysize = self.map_projection.round_up_pixels(
            (target_area.right - target_area.left) / self.map_projection.xstep,
            (target_area.top - target_area.bottom) / (self.map_projection.ystep * -1.0),
        )

        target_window = Window(xoff, yoff, xsize, ysize)
        return self._read_array_with_window(x, y, width, height, target_window)

    def _read_array(self, x: int, y: int, width: int, height: int) -> Any:
        return self._read_array_with_window(x, y, width, height, self.window)

    def read_array(self, x: int, y: int, width: int, height: int) -> Any:
        """Reads data from the layer based on the current reference window.

        Args:
            x: X axis offset for reading
            y: Y axis offset for reading
            width: Width of data to read
            height: Height of data to read

        Returns:
            An array of values from the layer.
        """
        res = self._read_array(x, y, width, height)
        return backend.demote_array(res)



from __future__ import annotations
import uuid
from enum import Enum
from typing import Any

from osgeo import gdal

from .._datatypes import Area, MapProjection, Window
from .rasters import YirgacheffeLayer, RasterLayer
from .._backends.enumeration import dtype as DataType

class VsimemFile:
    def __init__(self, path):
        self.path = path

    def __enter__(self):
        return self.path

    def __exit__(self, *args):
        try:
            gdal.Unlink(self.path)
        except RuntimeError:
            pass

class ResamplingMethod(Enum):
    """Resampling methods used in reprojecting rasters.

    This enumeration defines the resampling methods supported by Yirgacheffe.

    Attributes:
        Average: Computes the average of all non-NODATA contributing pixels
        Max: Selects the maximum value among all non-NODATA contributing pixels
        Med: Computes the median of all non-NODATA contributing pixels
        Min: Selects the minimum value among all non-NODATA contributing pixels
        Mode: Selects the most frequently occurring value among contributing pixels
        Nearest: Uses nearest-neighbor sampling (no interpolation)
        RootMeanSquare: Computes the root mean square of all non-NODATA contributing pixels
    """

    Average = "average"
    Max = "max"
    Med = "med"
    Min = "min"
    Mode = "mode"
    Nearest = "nearest"
    RootMeanSquare = "rms"
    Sum = "sum"

    # The commented out ones fail tests due to Yirgacheffe's chunking
    # behaviour and require more work.
    #
    # Bilinear = "bilinear"
    # Cubic = "cubic"
    # CubicSpline = "cubicspline"
    # Lanczos = "lanczos"
    # Q1 = "q1"
    # Q2 = "q2"

class ReprojectedRasterLayer(YirgacheffeLayer):
    """ReprojectedRasterLayer dynamically reprojects a layer.

    This layer wraps another layer and dynamically reprojects it to a different map projection
    and pixel scale. The reprojection is performed lazily - only when data is actually read.

    Thread Safety:
        Individual ReprojectedRasterLayer instances should not be accessed from multiple
        threads simultaneously. For parallel processing, use Python's multiprocessing module
        where each process creates its own layer instances. This is the recommended approach
        for raster processing as it avoids GIL limitations and ensures proper isolation.
    """

    def __init__(
        self,
        src: YirgacheffeLayer,
        target_projection: MapProjection,
        method: ResamplingMethod,
        name: str | None = None,
    ):
        # This calls `reset_window` on the layer as part of it's progress. In 2.0 this will be
        # a private API, but it's public in 1.x, though not widely used. This check is to ensure
        # that assumption about it not really being used is true.
        if src._active_area is not None:
            raise ValueError("Source can not have a custom window framing set")

        reprojected_area = src.area.reproject(target_projection)

        super().__init__(
            reprojected_area,
            name=name
        )

        self._src = src
        self._method = method

    @property
    def _cse_hash(self) -> int | None:
        return hash((
            self._src._cse_hash,
            self.name,
            self._underlying_area,
            self._method,
            self.map_projection,
            self._active_area
        ))

    def close(self):
        self._src.close()

    def _park(self):
        self._src._park()

    def _unpark(self):
        self._src._unpark()

    @property
    def datatype(self) -> DataType:
        return self._src.datatype

    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:

        # Step 1: work out what the area trying to be read is
        xoffset = xoffset + window.xoff
        yoffset = yoffset + window.yoff

        underlying_area = self._underlying_area
        projection = underlying_area.projection
        assert projection is not None
        read_area = Area(
            left=underlying_area.left + (xoffset * projection.xstep),
            top=underlying_area.top + (yoffset * projection.ystep),
            right=underlying_area.left + (xoffset * projection.xstep) + (xsize * projection.xstep),
            bottom=underlying_area.top + (yoffset * projection.ystep) + (ysize * projection.ystep),
            projection=projection,
        )

        # Resampling will generally require pixels outside the target area so we can say average correctly.
        # Mode resampling requires a larger buffer (3 pixels) to gather sufficient samples for determining
        # the most frequent value, whereas other methods just need a single pixel. Hopefully testing will
        # have sufficient coverage that we catch these as we expand the methods list.
        if self._method in {ResamplingMethod.Mode, ResamplingMethod.Sum}:
            expand_buffer = 3
        else:
            expand_buffer = 1

        # now we want this area in the source projection
        src_projection = self._src.map_projection
        assert src_projection is not None
        src_read_area = read_area.reproject(src_projection)

        # Note we should typically never go over the edge of the original
        # source material, as different resampling methods react differently to
        # the synthesised zeros that this will admit (e.g., one of min and max
        # would likely get confused).
        expanded_src_read_area = \
            src_read_area.grow(expand_buffer * src_projection.xstep)
        if self._method not in {ResamplingMethod.Sum}:
            expanded_src_read_area = expanded_src_read_area & self._src.area

        # We need some ID that stops us with other parallel workers potentially in the
        # VSIMEM space, so we use a uuid4 that should be close enough to collision free
        # without relying on things like pids.
        fid = uuid.uuid4()
        with VsimemFile(f"/vsimem/src_{fid}.tif") as src_data_path:
            # We need to be careful to restore the window state here, as the window is part of
            # the CSE hash for _src, and the CSE hash for this layer includes the hash for the _src
            # layer, so in effect we break our own hash if we forget to call reset window.
            if  self._src._active_area is not None:
                raise RuntimeError("Source can not have a custom window framing set")
            self._src._set_window(expanded_src_read_area)
            self._src.to_geotiff(src_data_path)
            self._src.reset_window()

            with VsimemFile(f"/vsimem/warped_{fid}.tif") as warped_data_path:
                gdal.Warp(
                    warped_data_path,
                    src_data_path,
                    options=gdal.WarpOptions(
                        dstSRS=projection._gdal_projection,
                        outputType=self.datatype.to_gdal(),
                        xRes=projection.xstep,
                        yRes=projection.ystep,
                        width=xsize,
                        height=ysize,
                        resampleAlg=self._method.value,
                        targetAlignedPixels=False,
                        outputBounds=(read_area.left, read_area.bottom, read_area.right, read_area.top)
                    )
                )

                with RasterLayer.layer_from_file(warped_data_path) as warped:
                    if (warped.window.xsize != xsize) or \
                        (warped.window.ysize != ysize):
                        raise RuntimeError(
                            f"gdal warp violated request constraints: "
                            f"expected {xsize}x{ysize}, got {warped.window.xsize}x{warped.window.ysize}"
                        )
                    return warped._read_array(0, 0, xsize, ysize)



from __future__ import annotations
import contextlib
import os
import tempfile
from multiprocessing import cpu_count
from pathlib import Path
from typing import Sequence

import numpy as np
from osgeo import gdal

from .layers import UniformAreaLayer
from .layers import YirgacheffeLayer
from .layers import ReprojectedRasterLayer, ResamplingMethod
from .layers import ConstantLayer
from .layers import GroupLayer, TiledGroupLayer
from .layers import RasterLayer
from .layers import VectorLayer
from .layers import AreaPerPixelLayer
from .window import MapProjection
from ._backends.enumeration import dtype as DataType

def read_raster(
    filename: Path | str,
    band: int = 1,
    ignore_nodata: bool = False,
) -> YirgacheffeLayer:
    """Open a raster file (e.g., GeoTIFF).

    Args:
        filename: Path of raster file to open.
        band: For multi-band rasters, which band to use (defaults to first if not specified).
        ignore_nodata: If the GeoTIFF has a NODATA value, don't substitute that value for NaN.

    Returns:
        An layer representing the raster data.

    Examples:
        >>> import yirgacheffe as yg
        >>> with yg.read_raster('test.tif') as layer:
        ...     total = layer.sum()
    """
    return RasterLayer.layer_from_file(filename, band, ignore_nodata)

def read_raster_like(
    filename: Path | str,
    like: YirgacheffeLayer,
    method: ResamplingMethod,
    band: int = 1,
    ignore_nodata: bool = False,
) -> YirgacheffeLayer:
    """Open a raster file but reproject it to match another layer.

    This method can be used to reproject a raster in one map projection to match another open layer.

    The reprojection will be only done when the data is actually used in a calculation, so unused data will
    not be reprojected.

    Args:
        filename: Path of raster file to open.
        like: Another layer that has a projection and pixel scale set. This layer will
            use the same projection and pixel scale as that one.
        method: The resampling method that will be used during the reprojection.
        band: For multi-band rasters, which band to use (defaults to first if not specified).
        ignore_nodata: If the GeoTIFF has a NODATA value, don't substitute that value for NaN.

    Returns:
        An layer representing the raster data in the new projection.

    Note:
        For parallel processing, use Python's multiprocessing module. This layer is safe when used
        with multiprocessing, but calls to `read_array` on an object are not thread safe.

    Examples:
        >>> import yirgacheffe as yg
        >>> with (
        ...     yg.read_raster('map_in_esri_54009.tif') as layer1,
        ...     yg.read_raster_like('map_in_wgs84.tif', layer1, yg.ResamplingMethod.Nearest) as layer2
        ... ):
        ...     res = layer1 * layer2
        ...     res.to_geotiff('result_in_esri_54009.tif')
    """
    if like.map_projection is None:
        raise ValueError("Reference layer must have a map projection.")
    original = RasterLayer.layer_from_file(filename, band, ignore_nodata)
    return ReprojectedRasterLayer(original, like.map_projection, method, original.name)

def read_narrow_raster(
    filename: Path | str,
    band: int = 1,
    ignore_nodata: bool = False,
) -> YirgacheffeLayer:
    """Open a 1 pixel wide raster file as a global raster.

    This exists for the special use case where an area per pixel raster would have the same value per horizontal row
    (e.g., a WGS84 map projection). For that case you can use this to load a raster that is 1 pixel wide and have
    it automatically expanded to act like a global raster in calculations.

    Args:
        filename: Path of raster file to open.
        band: For multi-band rasters, which band to use (defaults to first if not specified).
        ignore_nodata: If the GeoTIFF has a NODATA value, don't substitute that value for NaN.

    Returns:
        An layer representing the raster data.
    """
    return UniformAreaLayer.layer_from_file(filename, band, ignore_nodata)

def read_rasters(
    filenames : Sequence[Path | str],
    tiled: bool=False
) -> YirgacheffeLayer:
    """Open a set of raster files (e.g., GeoTIFFs) as a single layer.

    Args:
        filenames: List of paths of raster files to open.
        tiled: If you know that the rasters for a regular tileset, then setting this flag allows
            Yirgacheffe to perform certain optimisations that significantly improve performance for
            this use case.

    Returns:
        An layer representing the raster data.

    Examples:
        >>> import yirgacheffe as yg
        >>> with yg.read_rasters(['tile_N10_E10.tif', 'tile_N20_E10.tif']) as all_tiles:
        ...    ...
    """
    # Heading off a common error
    if isinstance(filenames, (str, Path)):
        raise TypeError(
            "Expected a sequence of files, not a single file. "
            f"Got {type(filenames).__name__}. Did you mean to pass [{filenames!r}]?"
        )
    if not tiled:
        return GroupLayer.layer_from_files(filenames)
    else:
        return TiledGroupLayer.layer_from_files(filenames)

def read_shape(
    filename: Path | str,
    projection: MapProjection | tuple[str, tuple[float, float]] | None = None,
    where_filter: str | None = None,
    datatype: DataType | None = None,
    burn_value: int | float | str = 1,
) -> YirgacheffeLayer:
    """Open a polygon file (e.g., GeoJSON, GPKG, or ESRI Shape File).

    Args:
        filename: Path of vector file to open.
        projection: The map projection to use.
        where_filter: For use with files with many entries (e.g., GPKG), applies this filter to the data.
        datatype: Specify the data type of the raster data generated.
        burn_value: The value of each pixel in the polygon.

    Returns:
        An layer representing the vector data.

    Examples:
        >>> import yirgacheffe as yg
        >>> with yg.read_shape('range.gpkg') as layer:
        ...    ...
    """

    if projection is not None:
        if not isinstance(projection, MapProjection):
            projection_name, scale_tuple = projection
            projection = MapProjection(projection_name, scale_tuple[0], scale_tuple[1])

    return VectorLayer._future_layer_from_file(
        filename,
        where_filter,
        projection,
        datatype,
        burn_value,
    )

def read_shape_like(
    filename: Path | str,
    like: YirgacheffeLayer,
    where_filter: str | None = None,
    datatype: DataType | None = None,
    burn_value: int | float | str = 1,
) -> YirgacheffeLayer:
    """Open a polygon file (e.g., GeoJSON, GPKG, or ESRI Shape File).

    Args:
        filename: Path of vector file to open.
        like: Another layer that has a projection and pixel scale set. This layer will
            use the same projection and pixel scale as that one.
        where_filter: For use with files with many entries (e.g., GPKG), applies this filter to the data.
        datatype: Specify the data type of the raster data generated.
        burn_value: The value of each pixel in the polygon.

    Returns:
        An layer representing the vector data.
    """
    return VectorLayer.layer_from_file_like(
        filename,
        like,
        where_filter,
        datatype,
        burn_value,
    )

def constant(value: int | float) -> YirgacheffeLayer:
    """Generate a layer that has the same value in all pixels regardless of scale, projection, and area.

    Generally this should not be necessary unless you must have the constant as the first term in an
    expression, as Yirgacheffe will automatically convert numbers into constant layers. However if the
    constant is the first term in the expression it must be wrapped by this call otherwise Python will
    not know that it should be part of the Yirgacheffe expression.

    Args:
        value: The value to be in each pixel of the expression term.

    Returns:
        A constant layer of the provided value.
    """
    return ConstantLayer(value)

def from_array(
    values: np.ndarray,
    origin: tuple[float, float],
    projection: MapProjection | tuple[str, tuple[float, float]],
) -> YirgacheffeLayer:
    """Creates an in-memory layer from a numerical array.

    Args:
        values: a 2D array of data values, with Y on the first dimension, X on
            the second dimension.
        origin: the position of the top left pixel in the geospatial space
        projection: the map projection and pixel scale to use.

    Returns:
        A geospatial layer that uses the provided data for its values.
    """

    if projection is None:
        raise ValueError("Projection must not be none")

    if not isinstance(projection, MapProjection):
        projection_name, scale_tuple = projection
        projection = MapProjection(projection_name, scale_tuple[0], scale_tuple[1])

    dims = values.shape

    # The original intent here was not to use GDAL directly, but use empty_raster_layer,
    # but then we end up tweaking the area multiple times, and hit rounding issues in
    # certain circumstances due to how it tries to align the output layer with a 0, 0
    # centered pixel grid.
    dataset = gdal.GetDriverByName('mem').Create(
        "mem",
        dims[1],
        dims[0],
        1,
        DataType.of_array(values).to_gdal(),
        [],
    )
    dataset.SetGeoTransform([
        origin[0], projection.xstep, 0.0, origin[1], 0.0, projection.ystep
    ])
    dataset.SetProjection(projection._gdal_projection)
    dataset.GetRasterBand(1).WriteArray(values, 0, 0)

    return RasterLayer(dataset)

def area_raster(
    projection: MapProjection | tuple[str, tuple[float, float]],
) -> YirgacheffeLayer:
    """Create a raster where the value of each pixel is the area in metres^2 of that pixel in the
    provided map projection and pixel scale.

    Args:
        projection: the map projection and pixel scale to use.

    Returns:
        A geospatial layer that where every pixel value is the geospatial area of that pixel.
    """

    if projection is None:
        raise ValueError("Projection must not be none")

    if not isinstance(projection, MapProjection):
        projection_name, scale_tuple = projection
        projection = MapProjection(projection_name, scale_tuple[0], scale_tuple[1])

    return AreaPerPixelLayer(projection)

def to_geotiff(
    filename: Path | str,
    bands: Sequence[YirgacheffeLayer],
    labels: list[str] | None = None,
    parallelism: int | bool | None = None,
    nodata: float | int | None = None,
    sparse: bool = False,
) -> None:
    """Save one or more results to a GeoTIFF file.

    Multiple results will be written as multiple bands in order. If labels are provided those will be
    assigned to the bands in the same order.

    All layers provided must be of the same map projection, pixel size, and data type. If necessary
    you can cast your layers using `astype` before passing them to this function.

    Args:
        filename: The name of the file to create.
        bands: A list of layers to store.
        labels: An optional list of band labels.
        parallelism: If passed, attempt to use multiple CPU cores up to the number provided, or if set to True,
            yirgacheffe will pick a sensible value.
        nodata: Nominate a value to be stored as nodata in the result.
        sparse: If True then save a sparse GeoTIFF as per GDAL's extension to the GeoTIFF standard.
    """
    if not bands:
        raise ValueError("Expected one or more layers to be written")

    if sparse and nodata is None:
        raise ValueError("Nodata value must be provided for sparse GeoTIFFs")

    layer_list = list(bands)
    first_layer = layer_list[0]
    for layer in layer_list[1:]:
        if layer.map_projection != first_layer.map_projection:
            raise ValueError("All layers must have the same map projection")
        if layer.datatype != first_layer.datatype:
            raise TypeError("All layers must have same data type. Use astype to explicitly cast layers.")

    if labels is not None and len(layer_list) != len(labels):
        raise ValueError("If labels are provided there must be the same number as there are layers")

    typed_filename = Path(filename)

    # We want to write to a tempfile before we move the result into place, but we can't use
    # the actual $TMPDIR as that might be on a different device, and so we use a file next to where
    # the final file will be, so we just need to rename the file at the end, not move it. But for special cases,
    # like GDAL's vsimem system we should not do this at all.
    target_dir_parts = typed_filename.parent.parts
    is_vsi_based = len(target_dir_parts) == 2 and target_dir_parts[0] =='/' and target_dir_parts[1] in [
        'vsimem', 'vsizip', 'vsigzip', 'vsi7z', 'vsirar', 'vsitar', 'vsistdin',
        'vsistdout', 'vsisubfile', 'vsiparse', 'vsicached', 'vsicrypt',
    ]
    if not is_vsi_based:
        typed_filename.parent.mkdir(parents=True, exist_ok=True)

    union_area = YirgacheffeLayer.find_union(layer_list)

    # GDAL TIFF compression is a significant bottleneck, and threading really helps. Yirgacheffe otherwise
    # will parallel compute a block of data, and then wait as a single thread does the GDAL TIFF compression
    # which is very, very slow without multithreading. Because they operate in turn, we can give all the cores
    # to GDAL as when it's active we're not using them and vice versa.
    gdal_tiff_threads = None
    if parallelism:
        if isinstance(parallelism, bool):
            gdal_tiff_threads = cpu_count()
        else:
            gdal_tiff_threads = parallelism

    projection = first_layer.map_projection
    if projection is None:
        raise ValueError("Can't save layours without projection")

    with tempfile.NamedTemporaryFile(dir=typed_filename.parent, delete=False) as tempory_file:
        try:
            with RasterLayer.empty_raster_layer(
                union_area,
                projection.scale,
                first_layer.datatype,
                tempory_file.name,
                projection.name,
                nodata=nodata,
                sparse=sparse,
                threads=gdal_tiff_threads,
                bands=len(layer_list),
            ) as output:
                for index, layer in enumerate(layer_list):
                    if parallelism is None:
                        _ = layer.save(output, band=index + 1)
                    else:
                        if isinstance(parallelism, bool):
                            # Parallel save treats None as "work it out"
                            parallelism = None
                        _ = layer.parallel_save(output, parallelism=parallelism, band=index + 1)
                    if labels:
                        output._dataset.GetRasterBand(index + 1).SetDescription(labels[index])

            if not is_vsi_based:
                os.rename(src=tempory_file.name, dst=typed_filename)
        except:
            with contextlib.suppress(OSError):
                os.unlink(tempory_file.name)
            raise



from __future__ import annotations
import sys
from dataclasses import dataclass

@dataclass(frozen=True)
class Window:
    """Class to hold the pixel dimensions of data in the given projection.

    Args:
        xoff: X axis offset.
        yoff: Y axis offset.
        xsize: Width of data in pixels.
        ysize: Height of data in pixels.

    Attributes:
        xoff: X axis offset.
        yoff: Y axis offset.
        xsize: Width of data in pixels.
        ysize: Height of data in pixels.
    """
    xoff: int
    yoff: int
    xsize: int
    ysize: int

    def __post_init__(self):
        for field in ['xoff', 'yoff', 'xsize', 'ysize']:
            value = getattr(self, field)
            if isinstance(value, bool):
                raise TypeError("Window properties must be whole numbers, not bool")
            if isinstance(value, float):
                if not value.is_integer():
                    raise ValueError("Window properties must be whole numbers, not bools")
                object.__setattr__(self, field, int(value))
            elif not isinstance(value, int):
                raise TypeError(
                    f"Window properties must be whole numbers, not {type(value).__name__}"
                )

    @property
    def as_array_args(self) -> tuple[int, ...]:
        """A tuple containing xoff, yoff, xsize, and ysize."""
        return (self.xoff, self.yoff, self.xsize, self.ysize)

    def __lt__(self, other) -> bool:
        return (self.xsize < other.xsize) and \
            (self.ysize < other.ysize) and \
            (self.xoff >= other.xoff) and \
            (self.yoff >= other.yoff) and \
            ((self.xoff + self.xsize) <= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) <= (other.yoff + other.ysize))

    def __gt__(self, other) -> bool:
        return (self.xsize > other.xsize) and \
            (self.ysize > other.ysize) and \
            (self.xoff <= other.xoff) and \
            (self.yoff <= other.yoff) and \
            ((self.xoff + self.xsize) >= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) >= (other.yoff + other.ysize))

    def __le__(self, other) -> bool:
        return (self.xsize <= other.xsize) and \
            (self.ysize <= other.ysize) and \
            (self.xoff >= other.xoff) and \
            (self.yoff >= other.yoff) and \
            ((self.xoff + self.xsize) <= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) <= (other.yoff + other.ysize))

    def __ge__(self, other) -> bool:
        return (self.xsize >= other.xsize) and \
            (self.ysize >= other.ysize) and \
            (self.xoff <= other.xoff) and \
            (self.yoff <= other.yoff) and \
            ((self.xoff + self.xsize) >= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) >= (other.yoff + other.ysize))

    def grow(self, pixels: int) -> Window:
        """Expand the area in all directions by the given amount.

        Generates a new window that is an expanded version of the current window.

        Args:
            pixels: The amount by which to grow the window in pixels.

        Returns:
            A new window of the expanded size.
        """
        return Window(
            xoff=self.xoff - pixels,
            yoff=self.yoff - pixels,
            xsize=self.xsize + (2 * pixels),
            ysize=self.ysize + (2 * pixels),
        )

    @staticmethod
    def find_intersection(windows: list) -> Window:
        if not windows:
            raise ValueError("Expected list of windows")
        # This isn't very pythonic, as it was originally written, but
        # this method gets called a lot (by every Layer.read_array), so not looping
        # over the window list multiple times halves the performance cost (as
        # measured with cProfile).
        left, top, right, bottom = -sys.maxsize, -sys.maxsize, sys.maxsize, sys.maxsize
        for window in windows:
            left = max(left, window.xoff)
            top = max(top, window.yoff)
            right = min(right, window.xoff + window.xsize)
            bottom = min(bottom, window.yoff + window.ysize)
        if (left >= right) or (top >= bottom):
            raise ValueError('No intersection possible')
        return Window(
            left,
            top,
            right - left,
            bottom - top
        )

    @staticmethod
    def find_intersection_no_throw(windows: list) -> Window | None:
        if not windows:
            raise ValueError("Expected list of windows")
        # This isn't very pythonic, as it was originally written, but
        # this method gets called a lot (by every Layer.read_array), so not looping
        # over the window list multiple times halves the performance cost (as
        # measured with cProfile).
        left, top, right, bottom = -sys.maxsize, -sys.maxsize, sys.maxsize, sys.maxsize
        for window in windows:
            left = max(left, window.xoff)
            top = max(top, window.yoff)
            right = min(right, window.xoff + window.xsize)
            bottom = min(bottom, window.yoff + window.ysize)
        if (left >= right) or (top >= bottom):
            return None
        return Window(
            left,
            top,
            right - left,
            bottom - top
        )




from .._backends import backend
from ..window import Window

class CSECacheTable:

    def __init__(self, expression, window: Window) -> None:
        self._table: dict[tuple[int, Window], tuple[int, backend.array_t | None]] = {}
        self._populate(expression, window)

    def __len__(self) -> int:
        return len(self._table)

    def _add(self, cse_hash: int | None, window: Window) -> int:
        if cse_hash is None:
            return 0

        try:
            count, data = self._table[(cse_hash, window)]
            cache_line = (count + 1, data)
        except KeyError:
            cache_line = (1, None)
        self._table[(cse_hash, window)] = cache_line
        count, _ = cache_line
        return count

    def _populate(self, expression, window: Window) -> None:
        used_window = window.grow(expression.buffer_padding)
        count = self._add(expression._cse_hash, used_window)
        if count == 1:
            # We only add children the first time we see an expression, otherwise
            # we will cache data covered by this node's cache line potentially
            for child in expression._children:
                try:
                    self._populate(child, used_window)
                except AttributeError:
                    try:
                        self._add(child._cse_hash, used_window)
                    except TypeError:
                        pass

    def set_data(self, cse_hash: int | None, window: Window, data: backend.array_t) -> None:
        if cse_hash is not None:
            try:
                count, old_data = self._table[(cse_hash, window)]
                if count < 2:
                    return
                if old_data is not None:
                    raise RuntimeWarning("Failure in CSE logic, setting data that is already set")
                self._table[(cse_hash, window)] = (count, data)
            except KeyError:
                raise RuntimeWarning("Failure in CSE logic, setting data for unknown term") # pylint: disable=W0707

    def get_data(self, cse_hash: int | None, window: Window) -> backend.array_t | None:
        # TODO: In theory we could release data from the table if we decremented the count each read
        # but we'd also need to fixed copy of the count for reset
        if cse_hash is None:
            return None
        try:
            _count, data = self._table[(cse_hash, window)]
            return data
        except KeyError:
            return None

    def reset_cache(self):
        self._table = {k: (count, None) for k, (count, _data) in self._table.items()}



from __future__ import annotations
from math import floor, ceil
from pathlib import Path
from typing import Any

from skimage import transform

from .._datatypes import Area, MapProjection, PixelScale, Window
from .rasters import RasterLayer, YirgacheffeLayer
from .._backends import backend
from .._backends.enumeration import dtype as DataType


class RescaledRasterLayer(YirgacheffeLayer):
    """RescaledRasterLayer dynamically rescales a raster, so to you to work with multiple layers at
    different scales without having to store unnecessary data. """

    @classmethod
    def layer_from_file(
        cls,
        filename: Path | str,
        pixel_scale: PixelScale,
        band: int = 1,
        nearest_neighbour: bool = True,
    ) -> RescaledRasterLayer:
        src = RasterLayer.layer_from_file(filename, band=band)
        source_projection = src.map_projection
        if source_projection is None:
            raise ValueError("Source raster must have projection and scale")
        target_projection = MapProjection(source_projection.name, pixel_scale.xstep, pixel_scale.ystep)
        return RescaledRasterLayer(src, target_projection, nearest_neighbour, src.name)

    def __init__(
        self,
        src: RasterLayer,
        target_projection: MapProjection,
        nearest_neighbour: bool = True,
        name: str | None = None,
    ):
        if src.map_projection is None:
            raise ValueError("Source layer must have a projection")
        if src.map_projection.name != target_projection.name:
            raise ValueError("Rescaled layer requires target projection is same.")

        target_area = Area(
            src.area.left,
            src.area.top,
            src.area.right,
            src.area.bottom,
            target_projection,
        )

        super().__init__(
            target_area,
            name=name
        )

        self._src = src
        self._nearest_neighbour = nearest_neighbour

        src_projection = src.map_projection

        self._x_scale = src_projection.xstep / target_projection.xstep
        self._y_scale = src_projection.ystep / target_projection.ystep

    @property
    def _cse_hash(self) -> int | None:
        return hash((
            self._src._cse_hash,
            self.name,
            self._underlying_area,
            self._nearest_neighbour,
            self.map_projection,
            self._active_area
        ))

    def close(self):
        self._src.close()

    def _park(self):
        self._src._park()

    def _unpark(self):
        self._src._unpark()

    @property
    def datatype(self) -> DataType:
        return self._src.datatype

    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:

        # to avoid aliasing issues, we try to scale to the nearest pixel
        # and recrop when scaling bigger

        xoffset = xoffset + window.xoff
        yoffset = yoffset + window.yoff

        src_x_offset = floor(xoffset / self._x_scale)
        src_y_offset = floor(yoffset / self._y_scale)

        diff_x = floor(((xoffset / self._x_scale) - src_x_offset) * self._x_scale)
        diff_y = floor(((yoffset / self._x_scale) - src_y_offset) * self._x_scale)

        src_x_width = ceil((xsize + diff_x) / self._x_scale)
        src_y_width = ceil((ysize + diff_y) / self._y_scale)

        # Get the matching src data
        src_data = backend.demote_array(self._src.read_array(
            src_x_offset,
            src_y_offset,
            src_x_width,
            src_y_width
        ))

        scaled = transform.resize(
            src_data,
            (src_y_width * self._y_scale, src_x_width * self._x_scale),
            order=(0 if self._nearest_neighbour else 1),
            anti_aliasing=(not self._nearest_neighbour)
        )

        return backend.promote(scaled[diff_y:(diff_y + ysize),diff_x:(diff_x + xsize)])



import pyproj

YSTEP = 512
MINIMUM_CHUNKS_PER_THREAD = 1

# Both GDAL and MLX assume that there is one instance running and it has the right to use
# all the memory it can. In general Yirgacheffe's chunking and it's own caching is what we shou
# be relying on, so we set some limits here. These are applied before each calculation, and so
# in theory they can be tweaked if necessary on demand.
GDAL_CACHE_LIMIT = 1 * 1024 * 1024 * 1024
MLX_CACHE_LIMIT = 1 * 1024 * 1024 * 1024

# I don't really want this here, but it's just too useful having it exposed
# This used to be a fixed string, but now it is at least programmatically generated
WGS_84_PROJECTION = pyproj.CRS.from_epsg(4326).to_wkt(version='WKT1_GDAL')



from __future__ import annotations

import builtins
import logging
import math
import multiprocessing
import operator as pyoperator
import os
import resource
import sys
import tempfile
import time
import types
from collections.abc import Callable
from contextlib import ExitStack, nullcontext, suppress
from enum import Enum
from functools import reduce
from multiprocessing import Process, Semaphore, cpu_count
from multiprocessing.synchronize import Semaphore as SemaphoreType
from multiprocessing.managers import SharedMemoryManager
from pathlib import Path
from typing import Any

import deprecation
import numpy as np
import numpy.typing as npt
from osgeo import gdal
from dill import dumps, loads # type: ignore
from pyproj import Transformer

from .. import constants, __version__
from .._datatypes import Area, PixelScale, MapProjection, Window
from .._backends import backend
from .._backends.enumeration import operators as op
from .._backends.enumeration import dtype as DataType
from .._backends.numpy import dtype_to_backend as dtype_to_numpy
from .._backends.numpy import backend_to_dtype as numpy_to_dtype
from .cse import CSECacheTable

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

class WindowOperation(Enum):
    NONE = 1
    UNION = 2
    INTERSECTION = 3

class LayerConstant:
    def __init__(self, val):
        self.val = val
        self.name = str(val)

    def __str__(self) -> str:
        return str(self.val)

    def _eval(self, _cse_cache, _area, _projection, _index, _step, _target_window):
        return self.val

    @property
    def _cse_hash(self) -> int | None:
        return hash(self.val)

    @property
    def datatype(self) -> DataType:
        numpy_type = np.result_type(self.val)
        return numpy_to_dtype(numpy_type)

    @property
    def area(self) -> Area:
        return Area.world()

    def _get_operation_area(self, _projection, _force_union, top_level) -> Area: # pylint: disable=W0613
        return Area.world()

class LayerMathMixin:

    def __add__(self, other):
        return LayerOperation(self, op.ADD, other, window_op=WindowOperation.UNION)

    def __radd__(self, other):
        return LayerOperation(self, op.RADD, other, window_op=WindowOperation.UNION)

    def __sub__(self, other):
        return LayerOperation(self, op.SUB, other, window_op=WindowOperation.UNION)

    def __rsub__(self, other):
        return LayerOperation(self, op.RSUB, other, window_op=WindowOperation.UNION)

    def __mul__(self, other):
        return LayerOperation(self, op.MUL, other, window_op=WindowOperation.INTERSECTION)

    def __rmul__(self, other):
        return LayerOperation(self, op.RMUL, other, window_op=WindowOperation.INTERSECTION)

    def __truediv__(self, other):
        return LayerOperation(self, op.TRUEDIV, other, window_op=WindowOperation.INTERSECTION)

    def __rtruediv__(self, other):
        return LayerOperation(self, op.RTRUEDIV, other, window_op=WindowOperation.INTERSECTION)

    def __floordiv__(self, other):
        return LayerOperation(self, op.FLOORDIV, other, window_op=WindowOperation.INTERSECTION)

    def __rfloordiv__(self, other):
        return LayerOperation(self, op.RFLOORDIV, other, window_op=WindowOperation.INTERSECTION)

    def __mod__(self, other):
        return LayerOperation(self, op.REMAINDER, other, window_op=WindowOperation.INTERSECTION)

    def __rmod__(self, other):
        return LayerOperation(self, op.RREMAINDER, other, window_op=WindowOperation.INTERSECTION)

    def __pow__(self, other):
        return LayerOperation(self, op.POW, other, window_op=WindowOperation.UNION)

    def __rpow__(self, other):
        return LayerOperation(self, op.RPOW, other, window_op=WindowOperation.UNION)

    def __eq__(self, other):
        return LayerOperation(self, op.EQ, other, window_op=WindowOperation.INTERSECTION)

    def __ne__(self, other):
        return LayerOperation(self, op.NE, other, window_op=WindowOperation.UNION)

    def __lt__(self, other):
        return LayerOperation(self, op.LT, other, window_op=WindowOperation.UNION)

    def __le__(self, other):
        return LayerOperation(self, op.LE, other, window_op=WindowOperation.UNION)

    def __gt__(self, other):
        return LayerOperation(self, op.GT, other, window_op=WindowOperation.UNION)

    def __ge__(self, other):
        return LayerOperation(self, op.GE, other, window_op=WindowOperation.UNION)

    def __and__(self, other):
        return LayerOperation(self, op.AND, other, window_op=WindowOperation.INTERSECTION)

    def __or__(self, other):
        return LayerOperation(self, op.OR, other, window_op=WindowOperation.UNION)

    def __lshift__(self, other):
        return LayerOperation(self, op.LSHIFT, other, window_op=WindowOperation.INTERSECTION)

    def __rshift__(self, other):
        return LayerOperation(self, op.RSHIFT, other, window_op=WindowOperation.INTERSECTION)

    def _eval(
        self,
        cse_cache: CSECacheTable,
        area: Area,
        projection: MapProjection,
        index: int,
        step: int,
        target_window: Window,
    ):
        cache_data = cse_cache.get_data(self._cse_hash, target_window)
        if cache_data is not None:
            return cache_data

        try:
            window = self.window if target_window is None else target_window
            result = self._read_array_for_area(area, projection, 0, index, window.xsize, step)
        except AttributeError:
            result = self._read_array_for_area(
                area,
                projection,
                0,
                index,
                target_window.xsize if target_window else 1,
                step
            )

        cse_cache.set_data(self._cse_hash, target_window, result)
        return result

    def nan_to_num(self, nan=0, posinf=None, neginf=None):
        return LayerOperation(
            self,
            op.NAN_TO_NUM,
            window_op=WindowOperation.NONE,
            copy=False,
            nan=nan,
            posinf=posinf,
            neginf=neginf,
        )

    def isin(self, test_elements):
        return LayerOperation(
            self,
            op.ISIN,
            window_op=WindowOperation.NONE,
            test_elements=tuple(test_elements),
        )

    def isnan(self):
        return LayerOperation(
            self,
            op.ISNAN,
            window_op=WindowOperation.NONE,
        )

    def abs(self):
        return LayerOperation(
            self,
            op.ABS,
            window_op=WindowOperation.NONE,
        )

    def __neg__(self):
        return LayerOperation(
            self,
            op.NEG,
            window_op=WindowOperation.NONE,
        )

    def __pos__(self):
        return LayerOperation(
            self,
            op.POS,
            window_op=WindowOperation.NONE,
        )

    def floor(self):
        return LayerOperation(
            self,
            op.FLOOR,
            window_op=WindowOperation.NONE,
        )

    def round(self):
        return LayerOperation(
            self,
            op.ROUND,
            window_op=WindowOperation.NONE,
        )

    def ceil(self):
        return LayerOperation(
            self,
            op.CEIL,
            window_op=WindowOperation.NONE,
        )

    def log(self):
        return LayerOperation(
            self,
            op.LOG,
            window_op=WindowOperation.NONE,
        )

    def log2(self):
        return LayerOperation(
            self,
            op.LOG2,
            window_op=WindowOperation.NONE,
        )

    def log10(self):
        return LayerOperation(
            self,
            op.LOG10,
            window_op=WindowOperation.NONE,
        )

    def exp(self):
        return LayerOperation(
            self,
            op.EXP,
            window_op=WindowOperation.NONE,
        )

    def exp2(self):
        return LayerOperation(
            self,
            op.EXP2,
            window_op=WindowOperation.NONE,
        )

    def clip(self, min=None, max=None): # pylint: disable=W0622
        # In the numpy 1 API np.clip(array) used a_max, a_min arguments and array.clip() used max and min as arguments
        # In numpy 2 they moved so that max and min worked on both, but still support a_max, and a_min on np.clip.
        # For now I'm only going to support the newer max/min everywhere notion, but I have to internally call
        # a_max, a_min so that yirgacheffe can work on older numpy installs.
        return LayerOperation(
            self,
            op.CLIP,
            window_op=WindowOperation.NONE,
            a_min=min,
            a_max=max,
        )

    def conv2d(self, weights):
        # A set of limitations that are just down to implementation time restrictions
        weights_width, weights_height = weights.shape
        if weights_width != weights_height:
            raise ValueError("Currently only square matrixes are supported for weights")
        padding = (weights_width - 1) / 2
        if not padding.is_integer():
            raise ValueError("Currently weights dimensions must be odd")

        return LayerOperation(
            self,
            op.CONV2D,
            window_op=WindowOperation.NONE,
            buffer_padding=int(padding),
            weights=weights.astype(np.float32),
        )

    def numpy_apply(self, func: Callable, other=None):
        return LayerOperation(self, func, other)

    def shader_apply(self, func: Callable, other=None):
        return ShaderStyleOperation(self, func, other)

    def save(self, destination_layer, and_sum=False, callback=None, band=1):
        return LayerOperation(self).save(destination_layer, and_sum, callback, band)

    def parallel_save(self, destination_layer, and_sum=False, callback=None, parallelism=None, band=1):
        return LayerOperation(self).parallel_save(destination_layer, and_sum, callback, parallelism, band)

    def parallel_sum(self, callback=None, parallelism=None, band=1):
        return LayerOperation(self).parallel_sum(callback, parallelism, band)

    def to_geotiff(
        self,
        filename: Path | str,
        and_sum: bool = False,
        parallelism: int | bool | None = None,
        callback: Callable[[float], None] | None = None,
        nodata: float | int | None = None,
        sparse: bool = False,
    ) -> float | None:
        """Saves a calculation to a raster file, optionally also returning the sum of pixels.

        Args:
            filename: Path of the raster to save the result to.
            and_sum: If true then the function will also calculate the sum of the raster as it goes and return
                that value.
            parallelism: If passed, attempt to use multiple CPU cores up to the number provided, or if set to True,
                yirgacheffe will pick a sensible value.
            callback: If passed, this callback will be called periodically with a progress update for the saving,
                with a value between 0.0 and 1.0.
            nodata: Nominate a value to be stored as nodata in the result.
            sparse: If True then save a sparse GeoTIFF as per GDAL's extension to the GeoTIFF standard.

        Returns:
            Either returns None, or the sum of the pixels in the resulting raster if `and_sum` was specified.
        """
        return LayerOperation(self).to_geotiff(
            filename=filename,
            and_sum=and_sum,
            parallelism=parallelism,
            callback=callback,
            nodata=nodata,
            sparse=sparse,
        )

    def sum(self):
        return LayerOperation(self).sum()

    def min(self):
        return LayerOperation(self).min()

    def max(self):
        return LayerOperation(self).max()

    def unique(self, return_counts=False):
        return LayerOperation(self).unique(return_counts)

    def astype(self, datatype):
        return LayerOperation(
            self,
            op.ASTYPE,
            window_op=WindowOperation.NONE,
            datatype=datatype
        )

    def latlng_for_pixel(self, x: int, y: int) -> tuple[float, float]:
        """Get geo coords for pixel. This is relative to the set view window.

        Args:
            x: X axis position within raster
            y: Y axis position within raster

        Returns:
            A tuple containing the (latitude, longitude).
        """
        projection = self.map_projection # type: ignore[attr-defined]
        area = self.area # type: ignore[attr-defined]
        if projection is None:
            raise ValueError("Map has not projection space")
        pixel_scale = projection.scale
        coord_in_raster_space = (
            (y * pixel_scale.ystep) + area.top,
            (x * pixel_scale.xstep) + area.left,
        )
        transformer = Transformer.from_crs(projection.name, "EPSG:4326")
        return transformer.transform(*coord_in_raster_space)

    def pixel_for_latlng(self, lat: float, lng: float) -> tuple[int, int]:
        """Get pixel for geo coords. This is relative to the set view window.
        Result is rounded down to nearest pixel.

        Args:
            lat: Geospatial latitude in WGS84
            lng: Geospatial longitude in WGS84

        Returns:
            A tuple containing the x, y coordinates in pixel space.
        """
        projection = self.map_projection # type: ignore[attr-defined]
        area = self.area # type: ignore[attr-defined]
        if projection is None:
            raise ValueError("Map has not projection space")

        transformer = Transformer.from_crs("EPSG:4326", projection.name)
        x, y = transformer.transform(lng,lat)

        return projection.round_down_pixels(
            (x - area.left) / projection.xstep,
            (y - area.top) / projection.ystep,
        )

    @property
    def window(self) -> Window:
        raise NotImplementedError("Must be overridden by subclass")

    @property
    def area(self) -> Area:
        raise NotImplementedError("Must be overridden by subclass")

    @property
    def _cse_hash(self) -> int | None:
        raise NotImplementedError("Must be overridden by subclass")

    def read_array(self, _x, _y, _w, _h):
        """Read an area of pixles from the specified area of a calculated raster.

        Args:
            x: X axis offset for reading
            y: Y axis offset for reading
            width: Width of data to read
            height: Height of data to read

        Returns:
            A numpy array containing the requested data. If the region of data read goes
            beyond the bounds of the calculation that area will be filled with zeros.
        """
        raise NotImplementedError("Must be overridden by subclass")

    def _read_array_for_area(self, _target_area, _target_projection, _x, _y, _w, _h):
        raise NotImplementedError("Must be overridden by subclass")

    def show(self, ax=None, max_pixels: int | None =1000, **kwargs):
        """Display data using matplotlib.

        Args:
            ax: Matplotlib axes object. If not provided, the default matplotlib context will be used.
            max_pixels: How many pixels to downsample to. If None, raw pixels will be used.
            **kwargs: Passed to matplotlib imshow.

        Returns:
            A matplotlib image.
        """
        import matplotlib.pyplot as plt # pylint: disable=C0415

        if ax is None:
            ax = plt.gca()

        window = self.window

        raw_data = self.read_array(
            0,
            0,
            window.xsize,
            window.ysize
        )
        if max_pixels:
            downsample = max(window.xsize, window.ysize) // max_pixels
            downsample = max(downsample, 1)
            data = raw_data[::downsample,::downsample]
        else:
            data = raw_data
        area = self.area
        extent = [
            area.left,
            area.right,
            area.bottom,
            area.top,
        ]
        return ax.imshow(data, extent=extent, **kwargs)


class LayerOperation(LayerMathMixin):

    @staticmethod
    @deprecation.deprecated(
        deprecated_in="1.10",
        removed_in="2.0",
        current_version=__version__,
        details="Use from top level module instead."
    )
    def where(cond, a, b):
        from . import functions # type: ignore # pylint: disable=C0415
        return functions.where(cond, a, b)

    @staticmethod
    @deprecation.deprecated(
        deprecated_in="1.10",
        removed_in="2.0",
        current_version=__version__,
        details="Use from top level module instead."
    )
    def minimum(a, b):
        from . import functions # type: ignore # pylint: disable=C0415
        return functions.minimum(a, b)

    @staticmethod
    @deprecation.deprecated(
        deprecated_in="1.10",
        removed_in="2.0",
        current_version=__version__,
        details="Use from top level module instead."
    )
    def maximum(a, b):
        from . import functions # type: ignore # pylint: disable=C0415
        return functions.maximum(a, b)

    def __init__(
        self,
        lhs: Any,
        operator: op | Callable | None = None,
        rhs: Any = None,
        other: Any = None,
        window_op: WindowOperation = WindowOperation.NONE,
        buffer_padding: int = 0,
        **kwargs
    ):
        self.ystep = constants.YSTEP
        self.kwargs = kwargs
        self.window_op = window_op
        self.buffer_padding = buffer_padding

        if lhs is None:
            raise ValueError("LHS on operation should not be none")
        self.lhs = lhs

        self.operator = operator

        if rhs is not None:
            if backend.isscalar(rhs):
                self.rhs: Any = LayerConstant(rhs)
            elif isinstance(rhs, (backend.array_t)):
                if rhs.shape == ():
                    self.rhs = LayerConstant(rhs.item())
                else:
                    raise ValueError("Numpy arrays are no allowed")
            else:
                if not ((lhs.map_projection == rhs.map_projection) or \
                        (lhs.map_projection is None) or \
                        (rhs.map_projection is None)):
                    raise ValueError("Not all layers are at the same pixel scale")
                self.rhs = rhs
        else:
            self.rhs = None

        if other is not None:
            if backend.isscalar(other):
                self.other: Any = LayerConstant(other)
            elif isinstance(other, (backend.array_t)):
                if other.shape == ():
                    self.rhs = LayerConstant(other.item())
                else:
                    raise ValueError("Numpy arrays are no allowed")
            else:
                if not ((lhs.map_projection == other.map_projection) or \
                        (lhs.map_projection is None) or \
                        (other.map_projection is None)):
                    raise ValueError("Not all layers are at the same pixel scale")
                self.other = other
        else:
            self.other = None

        # this is expensive, so don't do it all the time
        self._cse_hash_cache = self._calc_cse_hash()

    @property
    def _cse_hash(self) -> int | None:
        return self._cse_hash_cache

    def __str__(self) -> str:
        try:
            return f"({self.lhs} {self.operator} {self.rhs})"
        except AttributeError:
            try:
                return f"({self.operator} {self.lhs})"
            except AttributeError:
                return str(self.lhs)

    def __getstate__(self) -> object:
        odict = self.__dict__.copy()
        if isinstance(self.operator, types.LambdaType):
            odict['operator_dill'] = dumps(self.operator)
            del odict['operator']
        return odict

    def __setstate__(self, state) -> None:
        if 'operator_dill' in state:
            state['operator'] = loads(state['operator_dill'])
            del state['operator_dill']
        self.__dict__.update(state)

    @property
    def _children(self) -> list:
        return [x for x in [self.lhs, self.rhs, self.other] if x is not None]

    def _calc_cse_hash(self) -> int | None:
        # If we can't hash any of the child nodes then we can't hash this node, as if we can't store
        # their results in the cache, we can't know this result is stable
        child_hashes = [x._cse_hash for x in self._children]
        if any(x is None for x in child_hashes):
            return None

        # This really should be recursive
        def _make_hashable(value):
            if isinstance(value, (list, tuple, set)):
                return tuple(value)
            if isinstance(value, np.ndarray):
                return id(value)
            else:
                return value
        frozen_kwargs = tuple(sorted((k, _make_hashable(v)) for (k, v) in self.kwargs.items()))

        terms = [self.operator, self.window_op, frozen_kwargs, self.buffer_padding] + child_hashes

        try:
            return hash(tuple(terms))
        except TypeError:
            # This is assumed to be because kwargs contains something unhashable
            return None

    @property
    def area(self) -> Area:
        return self._get_operation_area(self.map_projection, top_level=True)

    def _get_operation_area(self, projection: MapProjection | None, force_union: bool = False, top_level=True) -> Area:
        lhs_area = self.lhs._get_operation_area(projection, force_union, top_level=False)
        try:
            rhs_area = self.rhs._get_operation_area(projection, force_union, top_level=False)
        except AttributeError:
            rhs_area = None
        try:
            other_area = self.other._get_operation_area(projection, force_union, top_level=False)
        except AttributeError:
            other_area = None

        all_areas = [x for x in [lhs_area, rhs_area, other_area] if x is not None]
        if force_union:
            all_areas = [x for x in all_areas if not x.is_world]

        window_op = WindowOperation.UNION if force_union else self.window_op
        match window_op:
            case WindowOperation.NONE:
                area = all_areas[0]
            case WindowOperation.INTERSECTION:
                area = reduce(pyoperator.and_, all_areas)
            case WindowOperation.UNION:
                area = reduce(pyoperator.or_, all_areas)
            case _:
                raise RuntimeError("Should not be reached")

        if top_level and area.is_world:
            return self._get_operation_area(projection, True)
        else:
            return area

    @property
    @deprecation.deprecated(
        deprecated_in="1.7",
        removed_in="2.0",
        current_version=__version__,
        details="Use `map_projection` instead."
    )
    def pixel_scale(self) -> PixelScale:
        # Because we test at construction that pixel scales for RHS/other are roughly equal,
        # I believe this should be sufficient...
        try:
            pixel_scale = self.lhs.pixel_scale
        except AttributeError:
            pixel_scale = None

        if pixel_scale is None:
            return self.rhs.pixel_scale
        return pixel_scale

    @property
    def window(self) -> Window:
        projection = self.map_projection
        if projection is None:
            # This can happen if your source layers are say just constants
            raise AttributeError("No window without projection")
        area = self._get_operation_area(projection)
        assert area is not None
        assert not area.is_world

        xoff, yoff = projection.round_down_pixels(
            area.left / projection.xstep,
            area.top / (projection.ystep * -1.0)
        )
        xsize, ysize = projection.round_up_pixels(
            (area.right - area.left) / projection.xstep,
            (area.top - area.bottom) / (projection.ystep * -1.0),
        )
        return Window(xoff, yoff, xsize, ysize)

    @property
    def datatype(self) -> DataType:
        # If this is an 'astype' then go with the target cast
        if self.operator == op.ASTYPE:
            try:
                return self.kwargs["datatype"]
            except KeyError as exc:
                raise RuntimeError("Internal inconsistency in astype operator") from exc
        # Otherwise we find the argument types and ask numpy to figure out what it would do
        internal_types: list[DataType] = [
            self.lhs.datatype
        ]
        if self.rhs is not None:
            internal_types.append(self.rhs.datatype)
        if self.other is not None:
            internal_types.append(self.other.datatype)
        internal_types_as_numpy_types = [dtype_to_numpy(x) for x in internal_types]
        coerced_type = np.result_type(*internal_types_as_numpy_types)
        return numpy_to_dtype(coerced_type)

    @property
    @deprecation.deprecated(
        deprecated_in="1.7",
        removed_in="2.0",
        current_version=__version__,
        details="Use `map_projection` instead."
    )
    def projection(self):
        try:
            projection = self.lhs.projection
        except AttributeError:
            projection = None

        if projection is None:
            projection = self.rhs.projection
        return projection

    @property
    def map_projection(self) -> MapProjection | None:
        try:
            projection = self.lhs.map_projection
        except AttributeError:
            projection = None

        if projection is None:
            try:
                projection = self.rhs.map_projection
            except AttributeError:
                pass
        return projection

    def pretty_print(self, prefix="", is_last=True):
        kwargs_str = ", ".join(f"{k}={v}" for k, v in self.kwargs.items())
        label = f"{self.operator}({kwargs_str})" if kwargs_str else self.operator
        label = f"{label} - {self._cse_hash}"

        connector = "└── " if is_last else "├── "
        print(f"{prefix}{connector}{label}")

        extension = "    " if is_last else "│   "
        new_prefix = prefix + extension

        children = self._children
        for i, child in enumerate(children):
            child_is_last = i == len(children) - 1
            if hasattr(child, 'pretty_print'):
                child.pretty_print(new_prefix, child_is_last)
            else:
                child_connector = "└── " if child_is_last else "├── "
                print(f"{new_prefix}{child_connector}{repr(child)}")

    def to_graphviz(self):
        """Create graphviz visualization showing shared nodes.

        Requires the graphviz python library is installed.

        Returns:
            A Graphviz object containing details of the expression.
        """
        import graphviz # pylint: disable=C0415
        dot = graphviz.Digraph(comment='Expression Tree')
        visited = set()
        self._add_to_graphviz(dot, visited)
        return dot

    def _add_to_graphviz(self, dot, visited):
        node_id = str(id(self))

        if node_id in visited:
            return  # Already added (shows sharing!)
        visited.add(node_id)

        # Create node label
        kwargs_str = ", ".join(f"{k}={v}" for k, v in self.kwargs.items())
        operator_label = str(self.operator).split(".")[1]
        label = f"{operator_label}\\n{kwargs_str}" if kwargs_str else str(operator_label)

        dot.node(node_id, label)

        # Add edges to children
        for _, child in enumerate(self._children):
            if hasattr(child, '_add_to_graphviz'):
                child_id = str(id(child))
                dot.edge(node_id, child_id)
                child._add_to_graphviz(dot, visited)
            else:
                leaf_id = f"leaf_{id(child)}"
                dot.node(leaf_id, child.name, shape='box')
                dot.edge(node_id, leaf_id)

    def _eval(
        self,
        cse_cache: CSECacheTable,
        area: Area,
        projection: MapProjection,
        index: int,
        step: int,
        target_window: Window,
    ):
        if self.buffer_padding:
            if target_window:
                target_window = target_window.grow(self.buffer_padding)
            area = area.grow(self.buffer_padding * projection.xstep)
            # The index doesn't need updating because we updated area/window
            step += (2 * self.buffer_padding)

        cache_data = cse_cache.get_data(self._cse_hash, target_window)
        if cache_data is not None:
            return cache_data

        lhs_data = self.lhs._eval(cse_cache, area, projection, index, step, target_window)

        if self.operator is None:
            result = lhs_data
        else:
            if isinstance(self.operator, op):
                operator = backend.operator_map[self.operator]
            else:
                # Handles things like `numpy_apply` where a custom operator is provided
                operator = self.operator

            if self.other is not None:
                assert self.rhs is not None
                rhs_data = self.rhs._eval(cse_cache, area, projection, index, step, target_window)
                other_data = self.other._eval(cse_cache, area, projection, index, step, target_window)
                result = operator(lhs_data, rhs_data, other_data, **self.kwargs)
            elif self.rhs is not None:
                rhs_data = self.rhs._eval(cse_cache, area, projection, index, step, target_window)
                result = operator(lhs_data, rhs_data, **self.kwargs)
            else:
                result = operator(lhs_data, **self.kwargs)

        cse_cache.set_data(self._cse_hash, target_window, result)
        return result

    def sum(self) -> float:
        # Numpy and MLX have different behaviours that make guessing
        # the type of the accumulator tricky, so we go by the type of the
        # first chunk returned
        res = None

        computation_window = self.window
        projection = self.map_projection
        if projection is None:
            raise ValueError("No map projection")

        cse_cache = CSECacheTable(self, computation_window)

        backend.init()

        for yoffset in range(0, computation_window.ysize, self.ystep):
            cse_cache.reset_cache()
            step = self.ystep
            if yoffset + step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(
                cse_cache,
                self._get_operation_area(projection),
                projection,
                yoffset,
                step,
                computation_window
            )
            chunk_sum = backend.sum_op(chunk)
            res = chunk_sum if res is None else res + chunk_sum
        return backend.demote_scalar(res) if res is not None else 0

    def min(self) -> float:
        res = float('inf')
        computation_window = self.window
        projection = self.map_projection
        if projection is None:
            raise ValueError("No map projection")

        cse_cache = CSECacheTable(self, computation_window)

        backend.init()

        for yoffset in range(0, computation_window.ysize, self.ystep):
            cse_cache.reset_cache()
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(
                cse_cache,
                self._get_operation_area(projection),
                projection,
                yoffset,
                step,
                computation_window
            )
            chunk_min = float(backend.min_op(chunk))
            res = min(res, chunk_min)
        return res

    def max(self) -> float:
        res = float('-inf')
        computation_window = self.window
        projection = self.map_projection
        if projection is None:
            raise ValueError("No map projection")

        cse_cache = CSECacheTable(self, computation_window)

        backend.init()

        for yoffset in range(0, computation_window.ysize, self.ystep):
            cse_cache.reset_cache()
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(
                cse_cache,
                self._get_operation_area(projection),
                projection,
                yoffset,
                step,
                computation_window
            )
            chunk_max = float(backend.max_op(chunk))
            res = max(res, chunk_max)
        return res

    def unique(self, return_counts:bool=False) -> np.ndarray | tuple[np.ndarray,np.ndarray]:
        """Returns a list of unique values found in the layer.

        Args:
            return_counts: If true, the function returns the counts of each item also.

        Returns:
            Either a list of unique items, or two lists, the first being the unique items
            and the second list being the corresponding counts.
        """
        res : dict[int|float,int] = {}
        computation_window = self.window
        projection = self.map_projection
        if projection is None:
            raise ValueError("No map projection")

        cse_cache = CSECacheTable(self, computation_window)

        backend.init()

        for yoffset in range(0, computation_window.ysize, self.ystep):
            cse_cache.reset_cache()
            step=self.ystep
            if yoffset+step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(
                cse_cache,
                self._get_operation_area(projection),
                projection,
                yoffset,
                step,
                computation_window
            )
            if return_counts:
                chunk_uniques, chunk_counts = backend.unique(chunk, return_counts=True)
            else:
                chunk_uniques = backend.unique(chunk)
                chunk_counts = [0] * len(chunk_uniques)
            for val, count in zip(chunk_uniques, chunk_counts):
                # Although you can set nan as a key in a dictionary, you can't access it
                # due to nan == nan being not true :) so we have to fudge it
                if np.isnan(val):
                    val = "nan"
                try:
                    res[val] = res[val] + count
                except KeyError:
                    res[val] = count
        if return_counts:
            uniques, counts = [], []
            for k, v in res.items():
                if k == "nan":
                    k = float("nan")
                uniques.append(k)
                counts.append(v)
            np_uniques = np.array(uniques)
            np_counts = np.array(counts)
            args = np.argsort(np_uniques)
            return np_uniques[args], np_counts[args]
        else:
            vals = [float("nan") if x == "nan" else x for x in res]
            return np.sort(np.array(vals))

    def save(self, destination_layer, and_sum=False, callback=None, band=1) -> float | None:
        """
        Calling save will write the output of the operation to the provied layer.
        If you provide sum as true it will additionall compute the sum and return that.
        """

        if destination_layer is None:
            raise ValueError("Layer is required")
        try:
            band = destination_layer._dataset.GetRasterBand(band)
        except AttributeError as exc:
            raise ValueError("Layer must be a raster backed layer") from exc

        projection = self.map_projection

        destination_window = destination_layer.window
        destination_projection = destination_layer.map_projection
        assert destination_projection is not None

        if projection is None:
            projection = destination_projection
        else:
            if projection != destination_projection:
                raise ValueError("Destination layer and input layers have different projection/scale")

        # If we're calculating purely from a constant layer, then we don't have a window or area
        # so we should use the destination raster details.
        try:
            computation_window = self.window
            computation_area = self._get_operation_area(projection)
        except (AttributeError, IndexError):
            computation_window = destination_window
            computation_area = destination_layer.area

        if (computation_window.xsize != destination_window.xsize) \
                or (computation_window.ysize != destination_window.ysize):
            computation_window = destination_window
            computation_area = destination_layer.area

        total = None

        cse_cache = CSECacheTable(self, computation_window)

        backend.init()

        for yoffset in range(0, computation_window.ysize, self.ystep):

            cse_cache.reset_cache()

            if callback:
                callback(yoffset / computation_window.ysize)
            step = self.ystep
            if yoffset + step > computation_window.ysize:
                step = computation_window.ysize - yoffset
            chunk = self._eval(cse_cache, computation_area, projection, yoffset, step, computation_window)
            if isinstance(chunk, (float, int)):
                chunk = backend.full((step, destination_window.xsize), chunk)
            band.WriteArray(
                backend.demote_array(chunk),
                destination_window.xoff,
                yoffset + destination_window.yoff,
            )
            if and_sum:
                chunk_sum = backend.sum_op(chunk)
                total = chunk_sum if total is None else total + chunk_sum
        if callback:
            callback(1.0)

        return backend.demote_scalar(total) if and_sum else None

    def _parallel_worker(
        self,
        index : int,
        shared_mem,
        sem : SemaphoreType,
        np_dtype : type,
        width : int,
        input_queue : multiprocessing.Queue,
        output_queue : multiprocessing.Queue,
        computation_window : Window,
    ):
        # The hashing of python objects isn't stable across processes in general, so we have to do
        # the cache build once per worker
        cse_cache = CSECacheTable(self, computation_window)

        backend.init()

        arr = np.ndarray((self.ystep, width), dtype=np_dtype, buffer=shared_mem.buf) # type: ignore[var-annotated]
        projection = self.map_projection
        # TODO: the `save` method does more sanity checking that parallel save!
        assert projection is not None
        try:
            while True:
                cse_cache.reset_cache()

                # We acquire the lock so we know we have somewhere to put the
                # result before we take work. This is because in practice
                # it seems the writing to GeoTIFF is the bottleneck, and
                # we had workers taking a task, then waiting for somewhere to
                # write to for ages when other workers were exiting because there
                # was nothing to do.
                sem.acquire()

                task = input_queue.get()
                if task is None:
                    sem.release()
                    output_queue.put(None)
                    break
                yoffset, step = task

                result = self._eval(
                    cse_cache,
                    self._get_operation_area(projection),
                    projection,
                    yoffset,
                    step,
                    computation_window,
                )
                backend.eval_op(result)

                arr[:step] = backend.demote_array(result)

                output_queue.put((index, yoffset, step))

        except Exception as e: # pylint: disable=W0718
            logger.exception(e)
            sem.release()
            output_queue.put(None)

    def _park(self):
        try:
            self.lhs._park()
        except AttributeError:
            pass
        try:
            self.rhs._park()
        except AttributeError:
            pass
        try:
            self.other._park()
        except AttributeError:
            pass

    def _parallel_save(
        self,
        destination_layer,
        and_sum=False,
        callback=None,
        parallelism=None,
        band=1
    ) -> float | None:
        assert (destination_layer is not None) or and_sum
        try:
            computation_window = self.window
        except (AttributeError, IndexError):
            # This is most likely because the calculation is on a constant layer (or combination of only constant
            # layers) and there's no real benefit to parallel saving then, so to keep this code from getting yet
            # more complicated just fall back to the single threaded path
            if destination_layer:
                return self.save(destination_layer, and_sum, callback, band)
            elif and_sum:
                return self.sum()
            else:
                raise RuntimeError("Should not be reached") # pylint: disable=W0707

        worker_count = parallelism or multiprocessing.cpu_count()
        work_blocks = len(range(0, computation_window.ysize, self.ystep))
        adjusted_blocks = math.ceil(work_blocks / constants.MINIMUM_CHUNKS_PER_THREAD)
        worker_count = min(adjusted_blocks, worker_count)

        if worker_count == 1:
            if destination_layer:
                return self.save(destination_layer, and_sum, callback, band)
            elif and_sum:
                return self.sum()
            else:
                raise RuntimeError("Should not be reached")

        if destination_layer is not None:
            try:
                gdal_band = destination_layer._dataset.GetRasterBand(band)
            except AttributeError as exc:
                raise ValueError("Layer must be a raster backed layer") from exc

            destination_window = destination_layer.window

            if (computation_window.xsize != destination_window.xsize) \
                    or (computation_window.ysize != destination_window.ysize):
                raise ValueError("Destination raster window size does not match input raster window size.")

            np_type_map: dict[int, np.dtype] = {
                gdal.GDT_Byte:    np.dtype('byte'),
                gdal.GDT_Float32: np.dtype('float32'),
                gdal.GDT_Float64: np.dtype('float64'),
                gdal.GDT_Int8:    np.dtype('int8'),
                gdal.GDT_Int16:   np.dtype('int16'),
                gdal.GDT_Int32:   np.dtype('int32'),
                gdal.GDT_Int64:   np.dtype('int64'),
                gdal.GDT_UInt16:  np.dtype('uint16'),
                gdal.GDT_UInt32:  np.dtype('uint32'),
                gdal.GDT_UInt64:  np.dtype('uint64'),
            }
            np_dtype = np_type_map[gdal_band.DataType]
        else:
            # the aggregation path
            gdal_band = None
            match self.datatype:
                case DataType.Int8 | DataType.Int16 | DataType.Int32 | DataType.Int64:
                    np_dtype = np.dtype('int64')
                case DataType.UInt8 | DataType.UInt16 | DataType.UInt32 | DataType.UInt64:
                    np_dtype = np.dtype('uint64')
                case _:
                    np_dtype = np.dtype('float64')

        # The parallel save will cause a fork on linux, so we need to
        # remove all SWIG references
        self._park()

        total = None

        with ExitStack() as stack:
            # If we get this far, then we're going to do the multiprocessing path. In general we've had
            # a lot of issues with limits on open file descriptors using multiprocessing on bigger machines
            # with hundreds of cores, and so to avoid blowing up in a way that is confusing to non-compsci
            # types, we just set the soft ulimit as high as we can
            previous_fd_limit, max_fd_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
            resource.setrlimit(resource.RLIMIT_NOFILE, (max_fd_limit, max_fd_limit))
            stack.callback(resource.setrlimit, resource.RLIMIT_NOFILE, (previous_fd_limit, max_fd_limit))

            with multiprocessing.Manager() as manager:
                with SharedMemoryManager() as smm:

                    mem_sem_cast = []
                    for _ in range(worker_count):
                        shared_buf = smm.SharedMemory(size=np_dtype.itemsize * self.ystep * computation_window.xsize)
                        cast_buf : npt.NDArray = np.ndarray(
                            (self.ystep, computation_window.xsize),
                            dtype=np_dtype,
                            buffer=shared_buf.buf
                        )
                        cast_buf[:] = np.zeros((self.ystep, computation_window.xsize), np_dtype)
                        mem_sem_cast.append((shared_buf, Semaphore(), cast_buf))

                    source_queue = manager.Queue()
                    result_queue = manager.Queue()

                    for yoffset in range(0, computation_window.ysize, self.ystep):
                        step = ((computation_window.ysize - yoffset)
                            if yoffset+self.ystep > computation_window.ysize
                            else self.ystep)
                        source_queue.put((
                            yoffset,
                            step
                        ))
                    for _ in range(worker_count):
                        source_queue.put(None)

                    if callback:
                        callback(0.0)

                    workers = [Process(target=self._parallel_worker, args=(
                        i,
                        mem_sem_cast[i][0],
                        mem_sem_cast[i][1],
                        np_dtype,
                        computation_window.xsize,
                        source_queue,
                        result_queue,
                        computation_window,
                    )) for i in range(worker_count)]
                    for worker in workers:
                        worker.start()

                    sentinal_count = len(workers)
                    retired_blocks = 0
                    while sentinal_count > 0:
                        res = result_queue.get()
                        if res is None:
                            sentinal_count -= 1
                            continue
                        index, yoffset, step = res
                        _, sem, arr = mem_sem_cast[index]
                        if gdal_band:
                            gdal_band.WriteArray(
                                arr[0:step],
                                destination_window.xoff,
                                yoffset + destination_window.yoff,
                            )
                        if and_sum:
                            chunk_sum = np.sum(np.array(arr[0:step]).astype(np.float64))
                            total = chunk_sum if total is None else total + chunk_sum
                        sem.release()
                        retired_blocks += 1
                        if callback:
                            callback(retired_blocks / work_blocks)

                    processes = workers
                    while processes:
                        candidates = [x for x in processes if not x.is_alive()]
                        for candidate in candidates:
                            candidate.join()
                            if candidate.exitcode:
                                for victim in processes:
                                    victim.kill()
                                sys.exit(candidate.exitcode)
                            processes.remove(candidate)
                        time.sleep(0.01)

        return backend.demote_scalar(total) if and_sum else None

    def parallel_save(
        self,
        destination_layer,
        and_sum=False,
        callback=None,
        parallelism=None,
        band=1
    ) -> float | None:
        if destination_layer is None:
            raise ValueError("Layer is required")
        return self._parallel_save(destination_layer, and_sum, callback, parallelism, band)

    def parallel_sum(self, callback=None, parallelism=None, band=1):
        return self._parallel_save(None, True, callback, parallelism, band)

    def to_geotiff(
        self,
        filename: Path | str,
        and_sum: bool = False,
        parallelism: int | bool | None = None,
        callback: Callable[[float], None] | None = None,
        nodata: float | int | None = None,
        sparse: bool = False,
    ) -> float | None:

        if sparse and nodata is None:
            raise ValueError("Nodata value must be provided for sparse GeoTIFFs")

        typed_filename = Path(filename)

        # We want to write to a tempfile before we move the result into place, but we can't use
        # the actual $TMPDIR as that might be on a different device, and so we use a file next to where
        # the final file will be, so we just need to rename the file at the end, not move it. But for special cases,
        # like GDAL's vsimem system we should not do this at all.
        target_dir_parts = typed_filename.parent.parts
        is_vsi_based = len(target_dir_parts) == 2 and target_dir_parts[0] =='/' and target_dir_parts[1] in [
            'vsimem', 'vsizip', 'vsigzip', 'vsi7z', 'vsirar', 'vsitar', 'vsistdin',
            'vsistdout', 'vsisubfile', 'vsiparse', 'vsicached', 'vsicrypt',
        ]
        if not is_vsi_based:
            typed_filename.parent.mkdir(parents=True, exist_ok=True)

        # GDAL TIFF compression is a significant bottleneck, and threading really helps. Yirgacheffe otherwise
        # will parallel compute a block of data, and then wait as a single thread does the GDAL TIFF compression
        # which is very, very slow without multithreading. Because they operate in turn, we can give all the cores
        # to GDAL as when it's active we're not using them and vice versa.
        gdal_tiff_threads = None
        if parallelism:
            if isinstance(parallelism, bool):
                gdal_tiff_threads = cpu_count()
            else:
                gdal_tiff_threads = parallelism

        context = (
            nullcontext(typed_filename) if is_vsi_based else
            tempfile.NamedTemporaryFile(dir=typed_filename.parent, delete=False)  # pylint: disable=R1732
        )
        with context as tempory_file:
            # Local import due to circular dependancy
            from yirgacheffe.layers.rasters import RasterLayer # type: ignore # pylint: disable=C0415
            inner_filename = tempory_file if is_vsi_based else tempory_file.name
            try:
                with RasterLayer.empty_raster_layer_like(
                    self,
                    filename=inner_filename, # type: ignore
                    nodata=nodata,
                    sparse=sparse,
                    threads=gdal_tiff_threads
                ) as layer:
                    if parallelism is None:
                        result = self.save(layer, and_sum=and_sum, callback=callback)
                    else:
                        if isinstance(parallelism, bool):
                            # Parallel save treats None as "work it out"
                            parallelism = None
                        result = self.parallel_save(layer, and_sum=and_sum, callback=callback, parallelism=parallelism)

                if not is_vsi_based:
                    os.rename(src=tempory_file.name, dst=filename)
            except:
                with suppress(OSError):
                    os.unlink(tempory_file.name)
                raise

        return result

    def read_array(self, x: int, y: int, width: int, height: int) -> np.ndarray:
        projection = self.map_projection
        if projection is None:
            raise ValueError("No map projection specified for layers in expression")

        computation_window = Window(0, 0, width, height)
        expression_area = self.area
        assert self.area.projection == projection
        left = expression_area.left + (x * projection.xstep)
        top = expression_area.top + (y * projection.ystep)
        computation_area = Area(
            left=left,
            top=top,
            right=left + (width * projection.xstep),
            bottom=top + (height * projection.ystep),
            projection=projection,
        )

        chunks = []

        cse_cache = CSECacheTable(self, computation_window)

        for yoffset in range(0, height, self.ystep):
            cse_cache.reset_cache()
            step = self.ystep
            if yoffset + step > height:
                step = height - yoffset
            chunk = self._eval(cse_cache, computation_area, projection, yoffset, step, computation_window)
            if isinstance(chunk, (float, int)):
                chunk = backend.full((step, computation_window.xsize), chunk)
            chunks.append(chunk)
        res = np.vstack(chunks)

        return res

    def _read_array_for_area(self, _target_area, _target_projection, _x, _y, _w, _h):
        raise RuntimeError("Should not be called")

class ShaderStyleOperation(LayerOperation):

    def _eval(self, cse_cache, area, projection, index, step, target_window=None):
        if target_window is None:
            target_window = self.window
        lhs_data = self.lhs._eval(cse_cache, area, projection, index, step, target_window)
        if self.rhs is not None:
            rhs_data = self.rhs._eval(cse_cache, area, projection, index, step, target_window)
        else:
            rhs_data = None

        # Constant results make this a bit messier. Might in future
        # be nicer to promote them to arrays sooner?
        if isinstance(lhs_data, (int, float)):
            if rhs_data is None:
                return self.operator(lhs_data, **self.kwargs)
            if isinstance(rhs_data, (int, float)):
                return self.operator(lhs_data, rhs_data, **self.kwargs)
            else:
                result = np.empty_like(rhs_data)
        else:
            result = np.empty_like(lhs_data)

        window = self.window
        for yoffset in range(step):
            for xoffset in range(window.xsize):
                try:
                    lhs_val = lhs_data[yoffset][xoffset]
                except TypeError:
                    lhs_val = lhs_data
                if rhs_data is not None:
                    try:
                        rhs_val = rhs_data[yoffset][xoffset]
                    except TypeError:
                        rhs_val = rhs_data
                    result[yoffset][xoffset] = self.operator(lhs_val, rhs_val, **self.kwargs)
                else:
                    result[yoffset][xoffset] = self.operator(lhs_val, **self.kwargs)

        return result

    def _read_array_for_area(self, _target_area, _target_projection, _x, _y, _w, _h):
        raise RuntimeError("Should not be called")



from __future__ import annotations

from enum import Enum

import numpy as np
from osgeo import gdal

class operators(Enum):
    ADD = 1
    SUB = 2
    MUL = 3
    TRUEDIV = 4
    POW = 5
    EQ = 6
    NE = 7
    LT = 8
    LE = 9
    GT = 10
    GE = 11
    AND = 12
    OR = 13
    LOG = 14
    LOG2 = 15
    LOG10 = 16
    EXP = 17
    EXP2 = 18
    CLIP = 19
    WHERE = 20
    MIN = 21
    MAX = 22
    SUM = 23
    MINIMUM = 24
    MAXIMUM = 25
    NAN_TO_NUM = 26
    ISIN = 27
    REMAINDER = 28
    FLOORDIV = 29
    CONV2D = 30
    ABS = 31
    ASTYPE = 32
    FLOOR = 33
    ROUND = 34
    CEIL = 35
    ISNAN = 36
    RADD = 37
    RSUB = 38
    RMUL = 39
    RTRUEDIV = 40
    RFLOORDIV = 41
    RREMAINDER = 42
    RPOW = 43
    UNIQUE = 44
    NEG = 45
    POS = 46
    LSHIFT = 47
    RSHIFT = 48

class dtype(Enum):
    """Represents the type of data returned by a layer.

    This enumeration defines the valid data types supported by Yirgacheffe, and is
    what is returned by  calling `datatype` on a layer or expression, and can be
    passed to `astype` to convert values between types.

    Attributes:
        Float32: 32 bit floating point value
        Float64: 64 bit floating point value
        Byte: Unsigned 8 bit integer value
        Int8: Signed 8 bit integer value
        Int16: Signed 16 bit integer value
        Int32: Signed 32 bit integer value
        Int64: Signed 64 bit integer value
        UInt8: Unsigned 8 bit integer value
        UInt16: Unsigned 16 bit integer value
        UInt32: Unsigned 32 bit integer value
        UInt64: Unsigned 64 bit integer value
    """

    Float32 = gdal.GDT_Float32
    Float64 = gdal.GDT_Float64
    Byte = gdal.GDT_Byte
    Int8 = gdal.GDT_Int8
    Int16 = gdal.GDT_Int16
    Int32 = gdal.GDT_Int32
    Int64 = gdal.GDT_Int64
    UInt8 = gdal.GDT_Byte
    UInt16 = gdal.GDT_UInt16
    UInt32 = gdal.GDT_UInt32
    UInt64 = gdal.GDT_UInt64

    def to_gdal(self) -> int:
        """Coverts the Yirgacheffe data type to the corresponding GDAL constant.

        Returns:
            An integer with the corresponding GDAL type constant.
        """
        return self.value

    @classmethod
    def of_gdal(cls, val: int) -> dtype:
        """Generates the Yirgacheffe data type value from the correspondiong GDAL value.

        Returns:
            A Yirgacheffe data type value.
        """
        return cls(val)

    @classmethod
    def of_array(cls, val: np.ndarray) -> dtype:
        """Generates the Yirgacheffe data type value from a numpy array.

        Returns:
            A Yirgacheffe data type value.
        """
        match val.dtype:
            case np.float32:
                return dtype.Float32
            case np.float64:
                return dtype.Float64
            case np.int8:
                return dtype.Int8
            case np.int16:
                return dtype.Int16
            case np.int32:
                return dtype.Int32
            case np.int64:
                return dtype.Int64
            case np.uint8:
                return dtype.UInt8
            case np.uint16:
                return dtype.UInt16
            case np.uint32:
                return dtype.UInt32
            case np.uint64:
                return dtype.UInt64
            case _:
                raise ValueError

    def sizeof(self) -> int:
        """Returns the number of bytes used to store the data type.

        Returns:
            The number of bytes used to store the data type.
        """
        match self:
            case dtype.Byte | dtype.Int8 | dtype.UInt8:
                return 1
            case dtype.Int16 | dtype.UInt16:
                return 2
            case dtype.Int32 | dtype.UInt32 | dtype.Float32:
                return 4
            case dtype.Int64 | dtype.UInt64 | dtype.Float64:
                return 8
            case _:
                raise ValueError
